from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, Type

from . import NetQASMInstruction, core, nv, trapped_ion, vanilla


@dataclass
class InstrMap:
    id_map: Optional[Dict[int, Type[NetQASMInstruction]]] = None
    name_map: Optional[Dict[str, Type[NetQASMInstruction]]] = None


CORE_INSTRUCTIONS: List[Type[NetQASMInstruction]] = [
    core.QAllocInstruction,
    core.InitInstruction,
    core.ArrayInstruction,
    core.SetInstruction,
    core.StoreInstruction,
    core.LoadInstruction,
    core.UndefInstruction,
    core.LeaInstruction,
    core.JmpInstruction,
    core.BezInstruction,
    core.BnzInstruction,
    core.BeqInstruction,
    core.BneInstruction,
    core.BltInstruction,
    core.BgeInstruction,
    core.AddInstruction,
    core.SubInstruction,
    core.AddmInstruction,
    core.SubmInstruction,
    core.MulInstruction,
    core.DivInstruction,
    core.RemInstruction,
    core.MeasInstruction,
    core.MeasBasisInstruction,
    core.CreateEPRInstruction,
    core.RecvEPRInstruction,
    core.WaitAllInstruction,
    core.WaitAnyInstruction,
    core.WaitSingleInstruction,
    core.QFreeInstruction,
    core.RetRegInstruction,
    core.RetArrInstruction,
    core.BreakpointInstruction,
]


class Flavour(ABC):
    """
    A Flavour represents an explicit instruction set that adheres to the Core NetQASM specification.
    Typically, a flavour is used for each specific target hardware.

    A flavour 'inherits' all classical instructions from the core, but can specify explicitly
    the quantum instructions that the hardware supports, by listing the corresponding instruction classes.

    Examples of flavours are the Vanilla flavour (with instructions defined in vanilla.py)
    and the Nitrogen-Vacancy (NV) flavour (instructions in nv.py).
    """

    def __init__(self, flavour_specific: List[Type[NetQASMInstruction]]):
        self.id_map = {instr.id: instr for instr in CORE_INSTRUCTIONS}
        self.id_map.update({instr.id: instr for instr in flavour_specific})

        self.name_map = {instr.mnemonic: instr for instr in CORE_INSTRUCTIONS}
        self.name_map.update({instr.mnemonic: instr for instr in flavour_specific})

    def get_instr_by_id(self, id: int):
        return self.id_map[id]

    def get_instr_by_name(self, name: str):
        return self.name_map[name]

    @property
    @abstractmethod
    def instrs(self):
        pass


class VanillaFlavour(Flavour):
    @property
    def instrs(self):
        return [
            vanilla.GateXInstruction,
            vanilla.GateYInstruction,
            vanilla.GateZInstruction,
            vanilla.GateHInstruction,
            vanilla.GateSInstruction,
            vanilla.GateKInstruction,
            vanilla.GateTInstruction,
            vanilla.RotXInstruction,
            vanilla.RotYInstruction,
            vanilla.RotZInstruction,
            vanilla.CnotInstruction,
            vanilla.CphaseInstruction,
            vanilla.MovInstruction,
        ]

    def __init__(self):
        super().__init__(self.instrs)


class NVFlavour(Flavour):
    @property
    def instrs(self):
        return [
            nv.RotXInstruction,
            nv.RotYInstruction,
            nv.RotZInstruction,
            nv.ControlledRotXInstruction,
            nv.ControlledRotYInstruction,
        ]

    def __init__(self):
        super().__init__(self.instrs)


class TrappedIonFlavour(Flavour):
    @property
    def instrs(self):
        return [
            trapped_ion.RotZInstruction,
            trapped_ion.AllQubitsRotXInstruction,
            trapped_ion.AllQubitsRotYInstruction,
            trapped_ion.AllQubitsRotZInstruction,
            trapped_ion.AllQubitsInitInstruction,
            trapped_ion.AllQubitsMeasInstruction,
            trapped_ion.BichromaticInstruction,
        ]

    def __init__(self):
        super().__init__(self.instrs)



"""
NetQASM subroutine definitions.

This module contains the `Subroutine` class which represents a static (not being
executed) NetQASM subroutine.
"""


from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

from netqasm.lang import encoding
from netqasm.lang.instr import DebugInstruction, NetQASMInstruction
from netqasm.lang.operand import Operand, Template
from netqasm.lang.version import NETQASM_VERSION
from netqasm.util.string import rspaces


class Subroutine:
    """
    A `Subroutine` object represents a subroutine consisting of valid
    instructions, i.e. objects deriving from `NetQASMInstruction`.

    Subroutines can optionally have *arguments*, which are yet-to-be-defined
    variables that are used in one or more of the instructions in the
    Subroutine. So, a Subroutine can be seen as a function which takes certain
    parameters (arguments). Concrete values for arguments can be given by
    instantiating (using the `instantiate` method).

    `Subroutine`s are executed by `Executor`s.
    """

    def __init__(
        self,
        instructions: Optional[List[NetQASMInstruction]] = None,
        arguments: Optional[List[str]] = None,
        netqasm_version: Tuple[int, int] = NETQASM_VERSION,
        app_id: Optional[int] = None,
    ) -> None:
        # TODO: remove?
        self._netqasm_version: Tuple[int, int] = netqasm_version
        # TODO: remove?
        self._app_id: Optional[int] = app_id

        self._instructions: List[NetQASMInstruction] = []
        if instructions is not None:
            self.instructions = instructions

        self._arguments: List[str] = []
        if arguments is not None:
            self._arguments = arguments
        else:
            # figure out argument by inspecting all commands
            for instr in self.instructions:
                for op in instr.operands:
                    if isinstance(op, Template):
                        self._arguments.append(op.name)

    @property
    def netqasm_version(self) -> Tuple[int, int]:
        return self._netqasm_version

    @property
    def app_id(self) -> Optional[int]:
        return self._app_id

    @app_id.setter
    def app_id(self, new_app_id: int) -> None:
        self._app_id = new_app_id

    @property
    def instructions(self) -> List[NetQASMInstruction]:
        return self._instructions

    @instructions.setter
    def instructions(self, new_instructions: List[NetQASMInstruction]) -> None:
        self._instructions = new_instructions

    @property
    def arguments(self) -> List[str]:
        return self._arguments

    def instantiate(
        self, app_id: int, arguments: Optional[Dict[str, int]] = None
    ) -> None:
        instrs: List[NetQASMInstruction] = []
        for instr in self.instructions:
            ops: List[Union[Operand, int]] = []
            for op in instr.operands:
                if isinstance(op, Template):
                    assert arguments is not None
                    ops.append(arguments[op.name])
                else:
                    ops.append(op)
            instrs.append(instr.from_operands(ops))

        self.instructions = instrs
        self._app_id = app_id
        if self._netqasm_version is None:
            self._netqasm_version = NETQASM_VERSION

    def pretty_print(self):
        result = "Subroutine"
        if len(self.arguments) > 0:
            result += "(" + ",".join(arg_name for arg_name in self.arguments) + ")"
        result += "\n"
        result += f"NetQASM version: {self.netqasm_version}\n"
        result += f"App ID: {self.app_id}\n"

        result += " LN | HLN | CMD\n"
        for i, instr in enumerate(self.instructions):
            if isinstance(instr, DebugInstruction):
                result += f"# {instr.text}\n"
            else:
                result += f"{rspaces(i)} {instr.debug_str}\n"
        result += "EndSubroutine"
        return result

    def print_instructions(self) -> str:
        return "\n".join(instr._pretty_print() for instr in self.instructions)

    def __str__(self):
        result = "Subroutine"
        if len(self.arguments) > 0:
            result += "(" + ",".join(arg_name for arg_name in self.arguments) + ")"
        result += "\n"
        for i, instr in enumerate(self.instructions):
            if isinstance(instr, DebugInstruction):
                result += f"# {instr.text}\n"
            else:
                result += f"  {instr._pretty_print()}\n"
        result += "EndSubroutine"
        return result

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Subroutine):
            return NotImplemented
        return (
            self.instructions == other.instructions
            and self.arguments == other.arguments
        )

    def __len__(self):
        return len(self.instructions)

    @property
    def cstructs(self):
        assert self.app_id is not None

        metadata = encoding.Metadata(
            netqasm_version=self.netqasm_version,
            app_id=self.app_id,
        )
        return [metadata] + [instr.serialize() for instr in self.instructions]

    def __bytes__(self):
        return b"".join(bytes(cstruct) for cstruct in self.cstructs)



from netqasm.examples.lib.bqc import (
    measXY,
    recv_meas_cmd,
    recv_teleported_state,
    send_meas_outcome,
)
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state

logger = get_netqasm_logger()


def main(app_config=None, num_iter=3):
    socket = Socket("server", "client", log_config=app_config.log_config)
    epr_socket = EPRSocket("client")

    num_qubits = num_iter + 1

    server = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
        max_qubits=num_qubits,
    )

    with server:
        # Receive qubits q[0] to q[num_qubits - 1] from the Client by teleportation.
        q = [None] * num_qubits
        for i in range(num_qubits):
            q[i] = recv_teleported_state(epr_socket)

        # Apply a CPHASE gate between every pair of consecutive qubits.
        for i in range(num_qubits - 1):
            q[i].cphase(q[i + 1])

        # TODO check why this is needed
        server.flush()

        # Main loop. Receive from the Client the angle to measure q[i] in.
        for i in range(num_iter):
            angle = recv_meas_cmd(socket)
            s = measXY(q[i], angle)
            server.flush()
            send_meas_outcome(socket, s)

        # The output of the computation is in the last qubit.
        dm = get_qubit_state(q[num_qubits - 1])
        return {"output_state": dm if dm is None else dm.tolist()}



"""
NetQASM execution interface for simulators.

This module provides the `Executor` class which can be used by simulators
as a base class for executing NetQASM instructions.
"""

from __future__ import annotations

import logging
import os
import traceback
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from itertools import count
from types import GeneratorType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import numpy as np
import qlink_interface as qlink_1_0

from netqasm.backend.network_stack import OK_FIELDS_K as OK_FIELDS
from netqasm.backend.network_stack import BaseNetworkStack
from netqasm.lang import instr as ins
from netqasm.lang import operand
from netqasm.lang.encoding import RegisterName
from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.operand import Address, ArrayEntry, ArraySlice
from netqasm.lang.parsing import parse_address
from netqasm.logging.glob import get_netqasm_logger
from netqasm.logging.output import InstrLogger
from netqasm.qlink_compat import (
    LinkLayerCreate,
    LinkLayerErr,
    LinkLayerOKTypeK,
    LinkLayerOKTypeM,
    LinkLayerOKTypeR,
    RequestType,
    ReturnType,
    get_creator_node_id,
    response_from_qlink_1_0,
)
from netqasm.sdk import shared_memory
from netqasm.sdk.shared_memory import Arrays, SharedMemory, SharedMemoryManager
from netqasm.util.error import NotAllocatedError

# Imports that are only needed for type checking
if TYPE_CHECKING:
    from netqasm.lang import subroutine as subrt_module

# Type definitions
T_UnitModule = List[Optional[int]]
T_LinkLayerResponse = Union[
    LinkLayerOKTypeK, LinkLayerOKTypeM, LinkLayerOKTypeR, LinkLayerErr
]
T_LinkLayerResponseOK = Union[LinkLayerOKTypeK, LinkLayerOKTypeM, LinkLayerOKTypeR]

T_RequestKey = Tuple[int, int]


@dataclass
class EprCmdData:
    """Container for info about EPR pending requests."""

    subroutine_id: int
    ent_results_array_address: int
    q_array_address: Optional[int]
    request: Optional[LinkLayerCreate]
    tot_pairs: int
    pairs_left: int


def inc_program_counter(method):
    """Decorator that automatically increases the current program counter.

    Should be used on functions that interpret a single NetQASM instruction.
    """

    def new_method(self, subroutine_id, instr):
        output = method(self, subroutine_id, instr)
        if isinstance(output, GeneratorType):
            output = yield from output
        self._program_counters[subroutine_id] += 1
        return output

    new_method.__name__ == method.__name__
    return new_method


class Executor:
    """Base class for entities that execute NetQASM applications.

    An Executor represents the component in a quantum node controller that handles
    the registration and execution of NetQASM applications.
    It can execute NetQASM subroutines by interpreting their instructions.

    This base class provides handlers for classical NetQASM instructions.
    These are methods with names `_instr_XXX`.
    Methods that handle quantum instructions are a no-op and should be overridden
    by subclasses.
    Entanglement instructions are handled and forwarded to the network stack.
    """

    # Global dictionary holding instruction logger objects per Executor instance
    _INSTR_LOGGERS: Dict[str, Optional[InstrLogger]] = {}

    # Class used for instruction loggers. May be different for subclasses of `Executor`.
    instr_logger_class = InstrLogger

    def __init__(
        self,
        name: Optional[str] = None,
        instr_log_dir: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Executor constructor.

        :param name: name of the executor for logging purposes, defaults to None
        :param instr_log_dir: directory to log instructions to, defaults to None
        """
        self._name: str  # declare type

        if name is None:
            self._name = f"{self.__class__}"
        else:
            self._name = name

        self._instruction_handlers: Dict[
            str, Callable
        ] = self._get_instruction_handlers()

        # Registers for different apps
        self._registers: Dict[int, Dict[RegisterName, shared_memory.RegisterGroup]] = {}

        # Arrays stored in memory for different apps
        self._app_arrays: Dict[int, Arrays] = {}

        # Shared memory with host for different apps
        self._shared_memories: Dict[int, SharedMemory] = {}

        self._qubit_unit_modules: Dict[int, T_UnitModule] = {}

        # There will be seperate program counters for each subroutine
        self._program_counters: Dict[int, int] = defaultdict(int)

        # Keep track of what subroutines are currently handled
        self._subroutines: Dict[int, subrt_module.Subroutine] = {}

        # Keep track of which subroutine in the order
        self._next_subroutine_id: int = 0

        # Keep track of what physical qubit addresses are in use
        self._used_physical_qubit_addresses: Set[int] = set()

        # Keep track of the create epr requests in progress
        self._epr_create_requests: Dict[T_RequestKey, List[EprCmdData]] = defaultdict(
            list
        )

        # Keep track of the recv epr requests in progress
        self._epr_recv_requests: Dict[T_RequestKey, List[EprCmdData]] = defaultdict(
            list
        )

        # Handle responsed for entanglement generation
        self._epr_response_handlers: Dict[
            ReturnType, Callable
        ] = self._get_epr_response_handlers()

        # Keep track of pending epr responses to handle
        self._pending_epr_responses: List[T_LinkLayerResponse] = []

        # Network stack
        self._network_stack: Optional[BaseNetworkStack] = None

        self._instr_logger: Optional[InstrLogger]  # declare type

        # Logger for instructions
        if instr_log_dir is None:
            self._instr_logger = None
        else:
            self._instr_logger = self.__class__.get_instr_logger(
                node_name=self._name,
                instr_log_dir=instr_log_dir,
                executor=self,
            )

        # Logger
        self._logger: logging.Logger = get_netqasm_logger(
            f"{self.__class__.__name__}({self._name})"
        )

    @property
    def name(self) -> str:
        """Get the name of this executor.

        :return: name
        """
        return self._name

    @property
    def node_id(self) -> int:
        """Get the ID of the node this Executor runs on

        :raises NotImplementedError: This should be overridden by a subclass
        :return: ID of the node
        """
        raise NotImplementedError

    def set_instr_logger(self, instr_log_dir: str) -> None:
        """Let the executor use an instruction logger that logs to `instr_log_dir`

        :param instr_log_dir: path to the log directory
        """
        self._instr_logger = self.__class__.get_instr_logger(
            node_name=self._name,
            instr_log_dir=instr_log_dir,
            executor=self,
            force_override=True,
        )

    @classmethod
    def get_instr_logger(
        cls,
        node_name: str,
        instr_log_dir: str,
        executor: Executor,
        force_override: bool = False,
    ) -> InstrLogger:
        instr_logger = cls._INSTR_LOGGERS.get(node_name)
        if instr_logger is None or force_override:
            filename = f"{str(node_name).lower()}_instrs.yaml"
            filepath = os.path.join(instr_log_dir, filename)
            instr_logger = cls.instr_logger_class(
                filepath=filepath,
                executor=executor,
            )
            cls._INSTR_LOGGERS[node_name] = instr_logger
        return instr_logger

    def _get_simulated_time(self) -> int:
        return 0

    @property
    def network_stack(self) -> Optional[BaseNetworkStack]:
        """Get the network stack (if any) connected to this Executor.

        :return: the network stack
        """
        return self._network_stack

    @network_stack.setter
    def network_stack(self, network_stack: BaseNetworkStack) -> None:
        if not isinstance(network_stack, BaseNetworkStack):
            raise TypeError(
                f"network_stack must be an instance of BaseNetworkStack, not {type(network_stack)}"
            )
        self._network_stack = network_stack

    def init_new_application(self, app_id: int, max_qubits: int) -> None:
        """Register a new application.

        :param app_id: App ID of the application.
        :param max_qubits: Maximum number of qubits the application is allowed to
            allocate at the same time.
        """
        self.allocate_new_qubit_unit_module(app_id=app_id, num_qubits=max_qubits)
        self._setup_registers(app_id=app_id)
        self._setup_arrays(app_id=app_id)
        self._new_shared_memory(app_id=app_id)

    def _setup_registers(self, app_id: int) -> None:
        """Setup registers for application"""
        self._registers[app_id] = shared_memory.setup_registers()

    def _setup_arrays(self, app_id: int) -> None:
        """Setup memory for storing arrays for application"""
        self._app_arrays[app_id] = Arrays()

    def _new_shared_memory(self, app_id: int) -> None:
        """Instantiate a new shared memory with an application"""
        self._shared_memories[app_id] = SharedMemoryManager.create_shared_memory(
            node_name=self._name, key=app_id
        )

    def setup_epr_socket(
        self, epr_socket_id: int, remote_node_id: int, remote_epr_socket_id: int
    ) -> Generator[Any, None, None]:
        """Instruct the Executor to open an EPR Socket.

        The Executor forwards this instruction to the Network Stack.

        :param epr_socket_id: ID of local EPR socket
        :param remote_node_id: ID of remote node
        :param remote_epr_socket_id: ID of remote EPR socket
        :yield: [description]
        """
        if self.network_stack is None:
            return
        output = self.network_stack.setup_epr_socket(
            epr_socket_id=epr_socket_id,
            remote_node_id=remote_node_id,
            remote_epr_socket_id=remote_epr_socket_id,
        )
        if isinstance(output, GeneratorType):
            yield from output

    def stop_application(self, app_id: int) -> Generator[Any, None, None]:
        """Stop an application and clear all qubits and classical memories.

        :param app_id: ID of the application to stop
        :yield: [description]
        """
        yield from self._clear_qubits(app_id=app_id)
        self._clear_registers(app_id=app_id)
        self._clear_arrays(app_id=app_id)
        self._clear_shared_memory(app_id=app_id)

    def _clear_qubits(self, app_id: int) -> Generator[Any, None, None]:
        unit_module = self._qubit_unit_modules.pop(app_id)
        for virtual_address, physical_address in enumerate(unit_module):
            if physical_address is None:
                continue
            self._used_physical_qubit_addresses.remove(physical_address)
            output = self._clear_phys_qubit_in_memory(physical_address)
            if isinstance(output, GeneratorType):
                yield from output

    def _clear_registers(self, app_id: int) -> None:
        self._registers.pop(app_id)

    def _clear_arrays(self, app_id: int) -> None:
        self._app_arrays.pop(app_id)

    def _clear_shared_memory(self, app_id: int) -> None:
        self._shared_memories.pop(app_id)

    def _reset_program_counter(self, subroutine_id: int) -> None:
        """Resets the program counter for a given subroutine ID"""
        self._program_counters.pop(subroutine_id, 0)

    def _clear_subroutine(self, subroutine_id: int) -> None:
        """Clears a subroutine from the executor"""
        self._reset_program_counter(subroutine_id=subroutine_id)
        self._subroutines.pop(subroutine_id, 0)

    def _get_instruction_handlers(self) -> Dict[str, Callable]:
        """Creates the dictionary of instruction handlers"""

        # For these core instructions, we provide a direct "_instr_{name}" method
        # The other types are handled in _execute_command
        mnemonic_mapping = [
            "qalloc",
            "array",
            "set",
            "store",
            "load",
            "undef",
            "lea",
            "meas",
            "create_epr",
            "recv_epr",
            "wait_all",
            "wait_any",
            "wait_single",
            "qfree",
            "ret_reg",
            "ret_arr",
        ]
        instruction_handlers = {
            mne: getattr(self, f"_instr_{mne}") for mne in mnemonic_mapping
        }
        return instruction_handlers

    def _get_epr_response_handlers(self) -> Dict[ReturnType, Callable]:
        """Get callbacks for EPR generation responses from the Network Stack.

        :return: dictionary of callbacks for each response type
        """
        epr_response_handlers = {
            ReturnType.ERR: self._handle_epr_err_response,
            ReturnType.OK_K: self._handle_epr_ok_k_response,
            ReturnType.OK_M: self._handle_epr_ok_m_response,
            ReturnType.OK_R: self._handle_epr_ok_r_response,
        }  # type: Dict[ReturnType, Callable]

        return epr_response_handlers

    def consume_execute_subroutine(self, subroutine: subrt_module.Subroutine) -> None:
        """Consume the generator returned by `execute_subroutine`.

        :param subroutine: subroutine to execute
        """
        list(self.execute_subroutine(subroutine=subroutine))

    def execute_subroutine(
        self, subroutine: subrt_module.Subroutine
    ) -> Generator[Any, None, None]:
        """Execute a NetQASM subroutine.

        This is a generator to allow simulators to yield at certain points during execution,
        e.g. to yield control to a asynchronous runtime.

        :param subroutine: subroutine to execute
        :yield: [description]
        """
        subroutine_id = self._get_new_subroutine_id()
        self._subroutines[subroutine_id] = subroutine
        self._reset_program_counter(subroutine_id)
        output = self._execute_commands(subroutine_id, subroutine.instructions)
        if isinstance(output, GeneratorType):
            yield from output
        self._clear_subroutine(subroutine_id=subroutine_id)

    def _get_new_subroutine_id(self) -> int:
        self._next_subroutine_id += 1
        return self._next_subroutine_id - 1

    def _execute_commands(
        self, subroutine_id: int, commands: List[NetQASMInstruction]
    ) -> Generator[Any, None, None]:
        """Execute a sequence of NetQASM instructions (commands) that are
        in a subroutine.

        :param subroutine_id: ID of the subroutine these instructions are in
        :param commands: list of NetQASM instructions
        :yield: [description]
        """
        while self._program_counters[subroutine_id] < len(commands):
            prog_counter = self._program_counters[subroutine_id]
            command = commands[prog_counter]
            try:
                output = self._execute_command(subroutine_id, command)
                if isinstance(
                    output, GeneratorType
                ):  # sanity check: should always be the case
                    yield from output
            except Exception as exc:
                traceback_str = "".join(traceback.format_tb(exc.__traceback__))
                self._handle_command_exception(exc, prog_counter, traceback_str)
                break

    def _handle_command_exception(
        self, exc: Exception, prog_counter: int, traceback_str: str
    ) -> None:
        raise exc.__class__(f"At line {prog_counter}: {exc}\n{traceback_str}") from exc

    def _execute_command(
        self, subroutine_id: int, command: NetQASMInstruction
    ) -> Generator[Any, None, None]:
        """Execute a single NetQASM instruction (command).

        :raises TypeError: if `command` is not a NetQASMInstruction
        :raises RuntimeError: if something went wrong while interpreting the
        instruction
        :yield: [description]
        """
        if not isinstance(command, NetQASMInstruction):
            raise TypeError(f"Expected a NetQASMInstruction, not {type(command)}")

        prog_counter = self._program_counters[subroutine_id]

        output = None
        if command.mnemonic in self._instruction_handlers:
            output = self._instruction_handlers[command.mnemonic](
                subroutine_id, command
            )
        else:
            if (
                isinstance(command, ins.core.SingleQubitInstruction)
                or isinstance(command, ins.core.InitInstruction)
                or isinstance(command, ins.core.QAllocInstruction)
                or isinstance(command, ins.core.QFreeInstruction)
            ):
                output = self._handle_single_qubit_instr(subroutine_id, command)
            elif isinstance(command, ins.core.TwoQubitInstruction):
                output = self._handle_two_qubit_instr(subroutine_id, command)
            elif isinstance(command, ins.core.RotationInstruction):
                output = self._handle_single_qubit_rotation(subroutine_id, command)
            elif isinstance(command, ins.core.ControlledRotationInstruction):
                output = self._handle_controlled_qubit_rotation(subroutine_id, command)
            elif (
                isinstance(command, ins.core.JmpInstruction)
                or isinstance(command, ins.core.BranchUnaryInstruction)
                or isinstance(command, ins.core.BranchBinaryInstruction)
            ):
                self._handle_branch_instr(subroutine_id, command)
            elif isinstance(command, ins.core.ClassicalOpInstruction) or isinstance(
                command, ins.core.ClassicalOpModInstruction
            ):
                output = self._handle_binary_classical_instr(subroutine_id, command)
            else:
                raise RuntimeError(f"unknown instr type: {type(command)}")

        if isinstance(output, GeneratorType):
            output = yield from output
        if self._instr_logger is not None:
            self._instr_logger.log(
                subroutine_id=subroutine_id,
                app_id=self._get_app_id(subroutine_id),
                command=command,
                output=output,
                program_counter=prog_counter,
            )

    @inc_program_counter
    def _instr_set(self, subroutine_id: int, instr: ins.core.SetInstruction) -> None:
        """Handle a NetQASM 'set' instruction."""
        self._logger.debug(f"Set register {instr.reg} to {instr.imm}")
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        self._set_register(app_id, instr.reg, instr.imm.value)

    def _set_register(
        self, app_id: int, register: operand.Register, value: int
    ) -> None:
        """Set the value of a register."""
        self._registers[app_id][register.name][register.index] = value

    def _get_register(self, app_id: int, register: operand.Register) -> Optional[int]:
        return self._registers[app_id][register.name][register.index]

    @inc_program_counter
    def _instr_qalloc(
        self, subroutine_id: int, instr: ins.core.QAllocInstruction
    ) -> int:
        """Handle a NetQASM `qalloc` instruction.

        :return: ID of physical qubit that got allocated
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        qubit_address = self._get_register(app_id, instr.reg)
        if qubit_address is None:
            raise RuntimeError(f"qubit address in register {instr.reg} is not defined")
        self._logger.debug(f"Taking qubit at address {qubit_address}")
        return self._allocate_physical_qubit(subroutine_id, qubit_address)

    @inc_program_counter
    def _instr_store(
        self, subroutine_id: int, instr: ins.core.StoreInstruction
    ) -> None:
        """Handle a NetQASM `store` instruction.

        Updates the relevant array entry.
        """
        register = instr.reg
        array_entry = instr.entry
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        value = self._get_register(app_id, register)
        if value is None:
            raise RuntimeError(f"value in register {register} is not defined")
        self._logger.debug(
            f"Storing value {value} from register {register} to array entry {array_entry}"
        )
        self._set_array_entry(app_id=app_id, array_entry=array_entry, value=value)

    @inc_program_counter
    def _instr_load(self, subroutine_id: int, instr: ins.core.LoadInstruction) -> None:
        """Handle a NetQASM `load` instruction.

        Loads the relevant array entry into a register.
        """
        register = instr.reg
        array_entry = instr.entry
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        value = self._get_array_entry(app_id=app_id, array_entry=array_entry)
        if value is None:
            raise RuntimeError(f"array value at {array_entry} is not defined")
        self._logger.debug(
            f"Storing value {value} from array entry {array_entry} to register {register}"
        )
        self._set_register(app_id, register, value)

    @inc_program_counter
    def _instr_lea(self, subroutine_id: int, instr: ins.core.LeaInstruction) -> None:
        """Handle a NetQASM `lea` instruction."""
        register = instr.reg
        address = instr.address
        self._logger.debug(f"Storing address of {address} to register {register}")
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        assert isinstance(address.address, int)
        self._set_register(app_id=app_id, register=register, value=address.address)

    @inc_program_counter
    def _instr_undef(
        self, subroutine_id: int, instr: ins.core.UndefInstruction
    ) -> None:
        """Handle a NetQASM `undef` instruction.

        Sets the relevant array entry to `None`.
        """
        array_entry = instr.entry
        self._logger.debug(f"Unset array entry {array_entry}")
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        self._set_array_entry(app_id=app_id, array_entry=array_entry, value=None)

    @inc_program_counter
    def _instr_array(
        self, subroutine_id: int, instr: ins.core.ArrayInstruction
    ) -> None:
        """Handle a NetQASM `array` instruction.

        Instantiates a new array with the relevant length.
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        length = self._get_register(app_id, instr.size)
        assert length is not None
        address = instr.address
        self._logger.debug(
            f"Initializing an array of length {length} at address {address}"
        )
        self._initialize_array(app_id=app_id, address=address, length=length)

    def _initialize_array(self, app_id: int, address: Address, length: int) -> None:
        arrays = self._app_arrays[app_id]
        assert isinstance(address.address, int)
        arrays.init_new_array(address.address, length)

    def _handle_branch_instr(
        self,
        subroutine_id: int,
        instr: Union[
            ins.core.BranchUnaryInstruction,
            ins.core.BranchBinaryInstruction,
            ins.core.JmpInstruction,
        ],
    ) -> None:
        """Handle a NetQASM branching instruction.

        The program counter is updated to the target label (in case of a jump)
        or to the next instruction (no jump).
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        a, b = None, None
        registers = []
        if isinstance(instr, ins.core.BranchUnaryInstruction):
            a = self._get_register(app_id=app_id, register=instr.reg)
            registers = [instr.reg]
        elif isinstance(instr, ins.core.BranchBinaryInstruction):
            a = self._get_register(app_id=app_id, register=instr.reg0)
            b = self._get_register(app_id=app_id, register=instr.reg1)
            registers = [instr.reg0, instr.reg1]

        if isinstance(instr, ins.core.JmpInstruction):
            condition = True
        elif isinstance(instr, ins.core.BranchUnaryInstruction):
            condition = instr.check_condition(a)
        elif isinstance(instr, ins.core.BranchBinaryInstruction):
            condition = instr.check_condition(a, b)

        if condition:
            jump_address = instr.line
            self._logger.debug(
                f"Branching to line {jump_address}, since {instr}(a={a}, b={b}) "
                f"is True, with values from registers {registers}"
            )
            self._program_counters[subroutine_id] = jump_address.value
        else:
            self._logger.debug(
                f"Don't branch, since {instr}(a={a}, b={b}) "
                f"is False, with values from registers {registers}"
            )
            self._program_counters[subroutine_id] += 1

    @inc_program_counter
    def _handle_binary_classical_instr(
        self,
        subroutine_id: int,
        instr: Union[
            ins.core.ClassicalOpInstruction,
            ins.core.ClassicalOpModInstruction,
        ],
    ) -> None:
        """Handle a NetQASM branching instruction with a binary condition."""
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        mod = None
        if isinstance(instr, ins.core.ClassicalOpModInstruction):
            mod = self._get_register(app_id=app_id, register=instr.regmod)
        if mod is not None and mod < 1:
            raise RuntimeError(f"Modulus needs to be greater or equal to 1, not {mod}")
        a = self._get_register(app_id=app_id, register=instr.regin0)
        b = self._get_register(app_id=app_id, register=instr.regin1)
        assert a is not None
        assert b is not None
        value = self._compute_binary_classical_instr(instr, a, b, mod=mod)
        mod_str = "" if mod is None else f"(mod {mod})"
        self._logger.debug(
            f"Performing {instr} of a={a} and b={b} {mod_str} "
            f"and storing the value {value} at register {instr.regout}"
        )
        self._set_register(app_id=app_id, register=instr.regout, value=value)

    def _compute_binary_classical_instr(
        self, instr: NetQASMInstruction, a: int, b: int, mod: Optional[int] = 1
    ):
        """Evaluate the binary condition of a NetQASM branching instruction."""
        if isinstance(instr, ins.core.AddInstruction):
            return a + b
        elif isinstance(instr, ins.core.AddmInstruction):
            assert mod is not None
            return (a + b) % mod
        elif isinstance(instr, ins.core.SubInstruction):
            return a - b
        elif isinstance(instr, ins.core.SubmInstruction):
            assert mod is not None
            return (a - b) % mod
        elif isinstance(instr, ins.core.MulInstruction):
            return a * b
        elif isinstance(instr, ins.core.DivInstruction):
            return a // b
        elif isinstance(instr, ins.core.RemInstruction):
            return a % b
        else:
            raise ValueError(f"{instr} cannot be used as binary classical function")

    @inc_program_counter
    def _handle_single_qubit_instr(
        self, subroutine_id: int, instr: ins.core.SingleQubitInstruction
    ) -> Generator[Any, None, None]:
        """Handle a single-qubit NetQASM instruction.

        Gets the virtual ID of the qubit and then calls `_do_single_qubit_instr`.

        This is a generator so that implementations of `_do_single_qubit_instr` can
        be generators themselves.
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address = self._get_register(app_id=app_id, register=instr.reg)
        assert q_address is not None
        self._logger.debug(f"Performing {instr} on the qubit at address {q_address}")
        output = self._do_single_qubit_instr(instr, subroutine_id, q_address)
        if isinstance(output, GeneratorType):
            yield from output

    def _do_single_qubit_instr(
        self, instr: ins.core.SingleQubitInstruction, subroutine_id: int, address: int
    ) -> Optional[Generator[Any, None, None]]:
        """Perform a single qubit gate on qubit with virtual ID `address`.

        This is a generator to allow simulators to yield at certain points during
        execution, e.g. to yield control to an asynchronous runtime.

        :param instr: NetQASM instruction to execute
        :param subroutine_id: ID of subroutine currently being executed
        :param address: virtual ID of qubit to act on
        :return: [description]
        """
        return None

    @inc_program_counter
    def _handle_single_qubit_rotation(
        self, subroutine_id: int, instr: ins.core.RotationInstruction
    ) -> Generator[Any, None, None]:
        """Handle a single-qubit NetQASM rotation instruction.

        Gets the virtual ID of the qubit and then calls `_do_single_qubit_rotation`.

        This is a generator so that implementations of `_do_single_qubit_rotation` can
        be generators themselves.
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address = self._get_register(app_id=app_id, register=instr.reg)
        assert q_address is not None
        angle = self._get_rotation_angle_from_operands(
            app_id=app_id,
            n=instr.angle_num.value,
            d=instr.angle_denom.value,
        )
        self._logger.debug(
            f"Performing {instr} with angle {angle} "
            f"on the qubit at address {q_address}"
        )
        output = self._do_single_qubit_rotation(
            instr, subroutine_id, q_address, angle=angle
        )
        if isinstance(output, GeneratorType):
            yield from output

    @inc_program_counter
    def _handle_controlled_qubit_rotation(
        self, subroutine_id: int, instr: ins.core.ControlledRotationInstruction
    ) -> Generator[Any, None, None]:
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address1 = self._get_register(app_id=app_id, register=instr.reg0)
        q_address2 = self._get_register(app_id=app_id, register=instr.reg1)
        assert q_address1 is not None
        assert q_address2 is not None
        angle = self._get_rotation_angle_from_operands(
            app_id=app_id, n=instr.angle_num.value, d=instr.angle_denom.value
        )
        self._logger.debug(
            f"Performing {instr} with angle {angle} "
            f"on the qubits at addresses {q_address1} and {q_address2}"
        )
        output = self._do_controlled_qubit_rotation(
            instr, subroutine_id, q_address1, q_address2, angle=angle
        )
        if isinstance(output, GeneratorType):
            yield from output

    def _get_rotation_angle_from_operands(self, app_id: int, n: int, d: int) -> float:
        return float(n * np.pi / 2**d)

    def _do_single_qubit_rotation(
        self,
        instr: ins.core.RotationInstruction,
        subroutine_id: int,
        address: int,
        angle: float,
    ) -> Optional[Generator[Any, None, None]]:
        """Perform a single-qubit rotation.

        This is a generator to allow simulators to yield at certain points during
        execution, e.g. to yield control to an asynchronous runtime.

        :param instr: NetQASM instruction to execute
        :param subroutine_id: ID of subroutine currently being executed
        :param address: virtual ID of qubit to act on
        :param angle: angle to rotate about
        :return: [description]
        """
        return None

    def _do_controlled_qubit_rotation(
        self,
        instr: ins.core.ControlledRotationInstruction,
        subroutine_id: int,
        address1: int,
        address2: int,
        angle: float,
    ) -> Optional[Generator[Any, None, None]]:
        """Perform a single-qubit controlled rotation.

        This is a generator to allow simulators to yield at certain points during
        execution, e.g. to yield control to an asynchronous runtime.

        :param instr: NetQASM instruction to execute
        :param subroutine_id: ID of subroutine currently being executed
        :param address1: virtual ID of control qubit
        :param address2: virtual ID of target qubit
        :param angle: angle to rotate about
        :return: [description]
        """
        return None

    @inc_program_counter
    def _handle_two_qubit_instr(
        self, subroutine_id: int, instr: ins.core.TwoQubitInstruction
    ) -> Generator[Any, None, None]:
        """Handle a two-qubit NetQASM instruction.

        Gets the virtual ID of the qubits and then calls `_do_two_qubit_instr`.

        This is a generator so that implementations of `_do_two_qubit_instr` can
        be generators themselves.
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address1 = self._get_register(app_id=app_id, register=instr.reg0)
        q_address2 = self._get_register(app_id=app_id, register=instr.reg1)
        assert q_address1 is not None
        assert q_address2 is not None
        self._logger.debug(
            f"Performing {instr} on the qubits at addresses {q_address1} and {q_address2}"
        )
        output = self._do_two_qubit_instr(instr, subroutine_id, q_address1, q_address2)
        if isinstance(output, GeneratorType):
            yield from output

    def _do_two_qubit_instr(
        self,
        instr: ins.core.TwoQubitInstruction,
        subroutine_id: int,
        address1: int,
        address2: int,
    ) -> Optional[Generator[Any, None, None]]:
        """Perform a two-qubit gate.

        This is a generator to allow simulators to yield at certain points during
        execution, e.g. to yield control to an asynchronous runtime.

        :param instr: NetQASM instruction to execute
        :param subroutine_id: ID of subroutine currently being executed
        :param address1: virtual ID of first qubit
        :param address2: virtual ID of second qubit
        :return: [description]
        """
        return None

    @inc_program_counter
    def _instr_meas(
        self, subroutine_id: int, instr: ins.core.MeasInstruction
    ) -> Generator[Any, None, None]:
        """Handle a `meas` NetQASM instruction.

        Gets the virtual ID of the qubit and then calls `_do_meas`.

        This is a generator so that implementations of `do_meas` can
        be generators themselves.

        :return: measurement outcome (0 or 1)
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address = self._get_register(app_id=app_id, register=instr.qreg)
        assert q_address is not None
        self._logger.debug(
            f"Measuring the qubit at address {q_address}, "
            f"placing the outcome in register {instr.creg}"
        )
        do_meas = self._do_meas(subroutine_id=subroutine_id, q_address=q_address)
        outcome: int
        if isinstance(do_meas, Generator):
            outcome = yield from do_meas  # type: ignore
        else:
            outcome = do_meas
        self._set_register(app_id=app_id, register=instr.creg, value=outcome)
        return outcome  # type: ignore

    def _do_meas(
        self, subroutine_id: int, q_address: int
    ) -> Union[int, Generator[int, None, None]]:
        """Perform a single-qubit measurement.

        This is a generator to allow simulators to yield at certain points during
        execution, e.g. to yield control to an asynchronous runtime.

        :param instr: NetQASM instruction to execute
        :param subroutine_id: ID of subroutine currently being executed
        :param address: virtual ID of qubit to act on
        :return: [description]
        """
        return 0

    @inc_program_counter
    def _instr_create_epr(
        self, subroutine_id: int, instr: ins.core.CreateEPRInstruction
    ) -> Generator[Any, None, None]:
        """Handle a `create_epr` NetQASM instruction.

        Gathers the relevant values from registers and then calls `_do_create_epr`.

        This is a generator so that implementations of `_do_create_epr` can
        be generators themselves.
        """
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        remote_node_id = self._get_register(
            app_id=app_id, register=instr.remote_node_id
        )
        epr_socket_id = self._get_register(app_id=app_id, register=instr.epr_socket_id)
        q_array_address = self._get_register(
            app_id=app_id, register=instr.qubit_addr_array
        )
        arg_array_address = self._get_register(app_id=app_id, register=instr.arg_array)
        ent_results_array_address = self._get_register(
            app_id=app_id, register=instr.ent_results_array
        )
        assert remote_node_id is not None
        assert epr_socket_id is not None
        # q_array_address can be None
        assert arg_array_address is not None
        assert ent_results_array_address is not None
        self._logger.debug(
            f"Creating EPR pair with remote node id {remote_node_id} and EPR socket ID {epr_socket_id}, "
            f"using qubit addresses stored in array with address {q_array_address}, "
            f"using arguments stored in array with address {arg_array_address}, "
            f"placing the entanglement information in array at address {ent_results_array_address}"
        )
        output = self._do_create_epr(
            subroutine_id=subroutine_id,
            remote_node_id=remote_node_id,
            epr_socket_id=epr_socket_id,
            q_array_address=q_array_address,
            arg_array_address=arg_array_address,
            ent_results_array_address=ent_results_array_address,
        )
        if isinstance(output, GeneratorType):
            yield from output

    def _do_create_epr(
        self,
        subroutine_id: int,
        remote_node_id: int,
        epr_socket_id: int,
        q_array_address: Optional[int],
        arg_array_address: int,
        ent_results_array_address: int,
    ) -> Optional[Generator[Any, None, None]]:
        """Send a request to the Network Stack to create EPR pairs.

        This is a generator so that implementations of `_do_single_qubit_rotation` can
        be generators themselves.

        :param remote_node_id: ID of remote node
        :param epr_socket_id: ID of local EPR socket
        :param q_array_address: address of array containing virtual qubits IDs.
        These IDs are mapped to local halves of the generated pairs.
        :param arg_array_address: address of array containing request information
        :param ent_results_array_address: address of array that will hold generation
        information after EPR generation has completed
        """
        if self.network_stack is None:
            raise RuntimeError("SubroutineHandler has no network stack")
        create_request = self._get_create_request(
            subroutine_id=subroutine_id,
            remote_node_id=remote_node_id,
            epr_socket_id=epr_socket_id,
            arg_array_address=arg_array_address,
        )
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        if create_request.type == RequestType.K:
            assert q_array_address is not None
            q_array = self._app_arrays[app_id][q_array_address, :]
            assert isinstance(q_array, list)
            num_qubits = len(q_array)
            assert num_qubits == create_request.number, "Not enough qubit addresses"
        self.network_stack.put(request=create_request)
        self._epr_create_requests[remote_node_id, create_request.purpose_id].append(
            EprCmdData(
                subroutine_id=subroutine_id,
                ent_results_array_address=ent_results_array_address,
                q_array_address=q_array_address,
                request=create_request,
                tot_pairs=create_request.number,
                pairs_left=create_request.number,
            )
        )
        return None

    def _get_create_request(
        self,
        subroutine_id: int,
        remote_node_id: int,
        epr_socket_id: int,
        arg_array_address: int,
    ) -> LinkLayerCreate:
        purpose_id = self._get_purpose_id(
            remote_node_id=remote_node_id, epr_socket_id=epr_socket_id
        )
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        array_args = self._app_arrays[app_id][arg_array_address, :]
        assert isinstance(array_args, list)

        # Not all values have to be defined (a default will be used).

        # assert all(arg is not None for arg in array_args)

        args = [remote_node_id, purpose_id] + array_args  # type: ignore

        # Use defaults if not specified
        expected_num_args = len(LinkLayerCreate._fields)
        if len(args) != expected_num_args:
            raise ValueError(
                f"Expected {expected_num_args} arguments, but got {len(args)}"
            )
        kwargs = {}
        for arg, field, default in zip(
            args, LinkLayerCreate._fields, LinkLayerCreate.__new__.__defaults__  # type: ignore
        ):
            if arg is None:
                kwargs[field] = default
            else:
                kwargs[field] = arg
        kwargs["type"] = RequestType(kwargs["type"])  # type: ignore

        return LinkLayerCreate(**kwargs)

    def _get_purpose_id(self, remote_node_id: int, epr_socket_id: int) -> int:
        # Should be subclassed
        # Note this is for now since we communicate directly to link layer
        if self._network_stack is None:
            raise RuntimeError("Exectioner has not network stack")
        return self._network_stack.get_purpose_id(
            remote_node_id=remote_node_id,
            epr_socket_id=epr_socket_id,
        )

    @inc_program_counter
    def _instr_recv_epr(
        self, subroutine_id: int, instr: ins.core.RecvEPRInstruction
    ) -> Generator[Any, None, None]:
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        remote_node_id = self._get_register(
            app_id=app_id, register=instr.remote_node_id
        )
        epr_socket_id = self._get_register(app_id=app_id, register=instr.epr_socket_id)
        q_array_address = self._get_register(
            app_id=app_id, register=instr.qubit_addr_array
        )
        ent_results_array_address = self._get_register(
            app_id=app_id, register=instr.ent_results_array
        )
        assert remote_node_id is not None
        assert epr_socket_id is not None
        # q_address can be None
        assert ent_results_array_address is not None
        self._logger.debug(
            f"Receiving EPR pair with remote node id {remote_node_id} "
            f"and EPR socket ID {epr_socket_id}, "
            f"using qubit addresses stored in array with address {q_array_address}, "
            f"placing the entanglement information in array at address {ent_results_array_address}"
        )
        output = self._do_recv_epr(
            subroutine_id=subroutine_id,
            remote_node_id=remote_node_id,
            epr_socket_id=epr_socket_id,
            q_array_address=q_array_address,
            ent_results_array_address=ent_results_array_address,
        )
        if isinstance(output, GeneratorType):
            yield from output

    def _do_recv_epr(
        self,
        subroutine_id: int,
        remote_node_id: int,
        epr_socket_id: int,
        q_array_address: Optional[int],
        ent_results_array_address: int,
    ) -> Optional[Generator[Any, None, None]]:
        if self.network_stack is None:
            raise RuntimeError("SubroutineHandler has no network stack")
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        # Get number of pairs based on length of ent info array
        num_pairs = self._get_num_pairs_from_array(
            app_id=app_id,
            ent_results_array_address=ent_results_array_address,
        )
        purpose_id = self._get_purpose_id(
            remote_node_id=remote_node_id,
            epr_socket_id=epr_socket_id,
        )
        self._epr_recv_requests[remote_node_id, purpose_id].append(
            EprCmdData(
                subroutine_id=subroutine_id,
                ent_results_array_address=ent_results_array_address,
                q_array_address=q_array_address,
                request=None,
                tot_pairs=num_pairs,
                pairs_left=num_pairs,
            )
        )
        return None

    def _get_num_pairs_from_array(
        self, app_id: int, ent_results_array_address: int
    ) -> int:
        ent_info = self._app_arrays[app_id][ent_results_array_address, :]
        assert isinstance(ent_info, list)
        return int(len(ent_info) / OK_FIELDS)

    @inc_program_counter
    def _instr_wait_all(
        self, subroutine_id: int, instr: ins.core.WaitAllInstruction
    ) -> Generator[Any, None, None]:
        array_slice = instr.slice
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        self._logger.debug(
            f"Waiting for all entries in array slice {array_slice} to become defined"
        )
        address, index = self._expand_array_part(app_id=app_id, array_part=array_slice)
        while True:
            values = self._app_arrays[app_id][address, index]
            if not isinstance(values, list):
                raise RuntimeError(
                    f"Something went wrong: values should be a list but it is a {type(values)}"
                )
            if any(value is None for value in values):
                output = self._do_wait()
                if isinstance(output, GeneratorType):
                    yield from output
            else:
                break
        self._logger.debug(f"Finished waiting for array slice {array_slice}")

    @inc_program_counter
    def _instr_wait_any(
        self, subroutine_id: int, instr: ins.core.WaitAnyInstruction
    ) -> Generator[Any, None, None]:
        array_slice = instr.slice
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        self._logger.debug(
            f"Waiting for any entry in array slice {array_slice} to become defined"
        )
        while True:
            values = self._get_array_slice(app_id=app_id, array_slice=array_slice)
            if values is None:
                raise RuntimeError(f"array slice {array_slice} does not exist")
            if all(value is None for value in values):
                output = self._do_wait()
                if isinstance(output, GeneratorType):
                    yield from output
            else:
                break
        self._logger.debug(f"Finished waiting for array slice {array_slice}")

    @inc_program_counter
    def _instr_wait_single(
        self, subroutine_id: int, instr: ins.core.WaitSingleInstruction
    ) -> Generator[Any, None, None]:
        array_entry = instr.entry
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        self._logger.debug(f"Waiting for array entry {array_entry} to become defined")
        while True:
            value = self._get_array_entry(app_id=app_id, array_entry=array_entry)
            if value is None:
                output = self._do_wait()
                if isinstance(output, GeneratorType):
                    yield from output
            else:
                break
        self._logger.debug(f"Finished waiting for array entry {array_entry}")

    def _do_wait(self) -> Optional[Generator[Any, None, None]]:
        return None

    @inc_program_counter
    def _instr_qfree(
        self, subroutine_id: int, instr: ins.core.QFreeInstruction
    ) -> Generator[Any, None, None]:
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        q_address = self._get_register(app_id=app_id, register=instr.reg)
        assert q_address is not None
        self._logger.debug(f"Freeing qubit at virtual address {q_address}")
        yield from self._free_physical_qubit(subroutine_id, q_address)

    @inc_program_counter
    def _instr_ret_reg(
        self, subroutine_id: int, instr: ins.core.RetRegInstruction
    ) -> None:
        register = instr.reg
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        value = self._get_register(app_id=app_id, register=register)
        if value is None:
            raise RuntimeError(
                f"Trying to return register {register} but it does not have value"
            )
        self._update_shared_memory(app_id=app_id, entry=register, value=value)

    @inc_program_counter
    def _instr_ret_arr(
        self, subroutine_id: int, instr: ins.core.RetArrInstruction
    ) -> None:
        address = instr.address
        app_id = self._get_app_id(subroutine_id=subroutine_id)

        array = self._get_array(app_id=app_id, address=address)

        # Not all values need to be defined.

        # if not all(elt is not None for elt in array):
        #     raise RuntimeError(
        #         f"Trying to return array {array} but not all values are defined yet"
        #     )

        self._update_shared_memory(
            app_id=app_id, entry=address, value=array  # type: ignore
        )

    def _update_shared_memory(
        self,
        app_id: int,
        entry: Union[operand.Register, Address, ArrayEntry, ArraySlice],
        value: Union[int, List[int]],
    ):
        shared_memory = self._shared_memories[app_id]
        if isinstance(entry, operand.Register):
            assert isinstance(value, int)
            self._logger.debug(
                f"Updating host about register {entry} with value {value}"
            )
            shared_memory.set_register(entry, value)
        elif isinstance(entry, ArrayEntry) or isinstance(entry, ArraySlice):
            self._logger.debug(
                f"Updating host about array entry {entry} with value {value}"
            )
            address, index = self._expand_array_part(app_id=app_id, array_part=entry)
            shared_memory.set_array_part(address=address, index=index, value=value)  # type: ignore
        elif isinstance(entry, Address):
            self._logger.debug(f"Updating host about array {entry} with value {value}")
            assert isinstance(entry.address, int)
            address = entry.address
            shared_memory.init_new_array(address=address, new_array=value)  # type: ignore
        else:
            raise TypeError(
                f"Cannot update shared memory with entry specified as {entry}"
            )

    def _get_unit_module(self, subroutine_id: int) -> T_UnitModule:
        app_id = self._get_app_id(subroutine_id)
        unit_module = self._qubit_unit_modules.get(app_id)
        if unit_module is None:
            raise RuntimeError(
                f"Application with app ID {app_id} has not allocated qubit unit module"
            )
        return unit_module

    def _get_position_in_unit_module(self, app_id: int, address: int) -> int:
        unit_module = self._qubit_unit_modules.get(app_id)
        if unit_module is None:
            raise RuntimeError(
                f"Application with app ID {app_id} has not allocated qubit unit module"
            )
        if address >= len(unit_module):
            raise IndexError(
                f"The address {address} is not within the allocated unit module "
                f"of size {len(unit_module)}"
            )
        position = unit_module[address]
        if position is None:
            raise NotAllocatedError(
                f"The qubit with address {address} was not allocated "
                f"for app ID {app_id} for node {self._name}"
            )
        return position

    def _get_array(self, app_id: int, address: Address) -> List[Optional[int]]:
        assert isinstance(address.address, int)
        return self._app_arrays[app_id]._get_array(address.address)

    def _get_array_entry(self, app_id: int, array_entry: ArrayEntry) -> Optional[int]:
        address, index = self._expand_array_part(app_id=app_id, array_part=array_entry)
        result = self._app_arrays[app_id][address, index]
        assert (result is None) or isinstance(result, int)
        return result

    def _set_array_entry(
        self, app_id: int, array_entry: ArrayEntry, value: Optional[int]
    ) -> None:
        address, index = self._expand_array_part(app_id=app_id, array_part=array_entry)
        self._app_arrays[app_id][address, index] = value

    def _get_array_slice(
        self, app_id: int, array_slice: ArraySlice
    ) -> Optional[List[Optional[int]]]:
        address, index = self._expand_array_part(app_id=app_id, array_part=array_slice)
        result = self._app_arrays[app_id][address, index]
        assert (result is None) or isinstance(result, list)
        return result

    def _expand_array_part(
        self, app_id: int, array_part: Union[ArrayEntry, ArraySlice]
    ) -> Tuple[int, Union[int, slice]]:
        address = array_part.address.address
        assert isinstance(address, int)
        index: Union[int, slice]
        if isinstance(array_part, ArrayEntry):
            if isinstance(array_part.index, int):
                index = array_part.index
            else:
                index_from_reg = self._get_register(
                    app_id=app_id, register=array_part.index
                )
                if index_from_reg is None:
                    raise RuntimeError(
                        f"Trying to use register {array_part.index} to index an array but its value is None"
                    )
                index = index_from_reg
        elif isinstance(array_part, ArraySlice):
            startstop: List[int] = []
            for raw_s in [array_part.start, array_part.stop]:
                if isinstance(raw_s, int):
                    startstop.append(raw_s)
                elif isinstance(raw_s, operand.Register):
                    s = self._get_register(app_id=app_id, register=raw_s)
                    if s is None:
                        raise RuntimeError(
                            f"Trying to use register {raw_s} to index an array but its value is None"
                        )
                    startstop.append(s)
                else:
                    raise RuntimeError(
                        f"Something went wrong: raw_s should be int or Register but is {type(raw_s)}"
                    )
            index = slice(*startstop)
        else:
            raise RuntimeError(
                f"Something went wrong: array_part is a {type(array_part)}"
            )
        return address, index

    def allocate_new_qubit_unit_module(self, app_id: int, num_qubits: int) -> None:
        unit_module = self._get_new_qubit_unit_module(num_qubits)
        self._qubit_unit_modules[app_id] = unit_module

    def _get_new_qubit_unit_module(self, num_qubits: int) -> T_UnitModule:
        return [None] * num_qubits

    def _has_virtual_address(self, app_id: int, virtual_address: int) -> bool:
        unit_module = self._qubit_unit_modules.get(app_id)
        if unit_module is None:
            return False
        if virtual_address < 0 or virtual_address >= len(unit_module):
            return False
        return unit_module[virtual_address] is not None

    def _allocate_physical_qubit(
        self,
        subroutine_id: int,
        virtual_address: int,
        physical_address: Optional[int] = None,
    ) -> int:
        """[summary]

        :param subroutine_id: [description]
        :param virtual_address: [description]
        :param physical_address: [description], defaults to None
        :raises ValueError: [description]
        :raises RuntimeError: [description]
        :return: physical qubit ID
        """
        unit_module = self._get_unit_module(subroutine_id)
        if virtual_address >= len(unit_module):
            app_id = self._subroutines[subroutine_id].app_id
            raise ValueError(
                f"Virtual address {virtual_address} is outside the unit module (app ID {app_id}) "
                f"which has length {len(unit_module)}"
            )
        if unit_module[virtual_address] is None:
            if physical_address is None:
                physical_address = self._get_unused_physical_qubit()
                self._used_physical_qubit_addresses.add(physical_address)
            unit_module[virtual_address] = physical_address
            self._reserve_physical_qubit(physical_address)
            return physical_address
        else:
            app_id = self._subroutines[subroutine_id].app_id
            raise RuntimeError(
                f"QubitAddress at address {virtual_address} "
                f"for application {app_id} is already allocated"
            )

    def _free_physical_qubit(
        self, subroutine_id: int, address: int
    ) -> Generator[Any, None, None]:
        unit_module = self._get_unit_module(subroutine_id)
        if unit_module[address] is None:
            app_id = self._subroutines[subroutine_id].app_id
            raise RuntimeError(
                f"QubitAddress at address {address} for application {app_id} is not allocated "
                "and cannot be freed"
            )
        else:
            physical_address = unit_module[address]
            assert physical_address is not None
            self._logger.debug(f"Freeing qubit at physical address {physical_address}")
            unit_module[address] = None
            self._used_physical_qubit_addresses.remove(physical_address)
            output = self._clear_phys_qubit_in_memory(physical_address)
            if isinstance(output, GeneratorType):
                yield from output

    def _reserve_physical_qubit(
        self, physical_address: int
    ) -> Generator[Any, None, None]:
        """To be subclassed for different quantum processors (e.g. netsquid)"""
        yield None

    def _clear_phys_qubit_in_memory(
        self, physical_address: int
    ) -> Generator[Any, None, None]:
        """To be subclassed for different quantum processors (e.g. netsquid)"""
        yield None

    def _get_unused_physical_qubit(self) -> int:
        # Assuming that the topology of the unit module is a complete graph
        # is does not matter which unused physical qubit we choose for now
        for physical_address in count(0):
            if physical_address not in self._used_physical_qubit_addresses:
                self._used_physical_qubit_addresses.add(physical_address)
                return physical_address
        raise RuntimeError("should never get here")

    def _get_app_id(self, subroutine_id: int) -> int:
        """Returns the app ID for the given subroutine"""
        subroutine = self._subroutines.get(subroutine_id)
        if subroutine is None:
            raise ValueError(f"Unknown subroutine with ID {subroutine_id}")
        assert subroutine.app_id is not None
        return subroutine.app_id

    def _handle_epr_response(self, response: T_LinkLayerResponse) -> None:

        if (
            isinstance(response, qlink_1_0.ResCreateAndKeep)
            or isinstance(response, qlink_1_0.ResMeasureDirectly)
            or isinstance(response, qlink_1_0.ResError)
        ):
            # Convert from qlink-layer 1.0
            response = response_from_qlink_1_0(response)

        self._pending_epr_responses.append(response)
        self._handle_pending_epr_responses()

    def _handle_pending_epr_responses(self) -> None:
        # NOTE this will probably be handled differently in an actual implementation
        # but is done in a simple way for now to allow for simulation
        if len(self._pending_epr_responses) == 0:
            return

        # Try to handle one of the pending EPR responses
        handled = False
        for i, response in enumerate(self._pending_epr_responses):

            if response.type == ReturnType.ERR:
                self._handle_epr_err_response(response)  # type: ignore
            else:
                self._logger.debug(
                    f"Try to handle EPR OK ({response.type}) response from network stack"
                )
                info = self._extract_epr_info(response=response)  # type: ignore
                if info is not None:
                    epr_cmd_data, pair_index, is_creator, request_key = info
                    handled = self._epr_response_handlers[response.type](
                        epr_cmd_data=epr_cmd_data,
                        response=response,
                        pair_index=pair_index,
                    )
                if handled:
                    epr_cmd_data.pairs_left -= 1

                    self._handle_last_epr_pair(
                        epr_cmd_data=epr_cmd_data,
                        is_creator=is_creator,
                        request_key=request_key,
                    )

                    self._store_ent_info(
                        epr_cmd_data=epr_cmd_data,
                        response=response,  # type: ignore
                        pair_index=pair_index,
                    )
                    self._pending_epr_responses.pop(i)
                    break
        if not handled:
            self._wait_to_handle_epr_responses()
        else:
            self._handle_pending_epr_responses()

    def _wait_to_handle_epr_responses(self) -> None:
        # This can be subclassed to sleep a little before handling again
        self._handle_pending_epr_responses()

    def _handle_epr_err_response(self, response: LinkLayerErr) -> None:
        raise RuntimeError(
            f"Got the following error from the network stack: {response}"
        )

    def _extract_epr_info(
        self, response: T_LinkLayerResponseOK
    ) -> Optional[Tuple[EprCmdData, int, bool, T_RequestKey]]:
        creator_node_id: int = get_creator_node_id(self.node_id, response)  # type: ignore

        # Retreive the data for this request (depending on if we are creator or receiver
        if creator_node_id == self.node_id:
            is_creator = True
            requests = self._epr_create_requests
        else:
            is_creator = False
            requests = self._epr_recv_requests

        purpose_id = response.purpose_id
        remote_node_id = response.remote_node_id
        request_key = remote_node_id, purpose_id
        if len(requests[request_key]) == 0:
            self._logger.debug(
                f"Since there is yet not recv request for remote node ID {remote_node_id} and "
                f"purpose ID {purpose_id}, "
                "handling of epr will wait and try again."
            )
            return None
        epr_cmd_data = requests[request_key][0]

        pair_index = epr_cmd_data.tot_pairs - epr_cmd_data.pairs_left

        return epr_cmd_data, pair_index, is_creator, request_key

    def _handle_last_epr_pair(
        self, epr_cmd_data: EprCmdData, is_creator: bool, request_key: T_RequestKey
    ) -> None:
        # Check if this was the last pair
        if epr_cmd_data.pairs_left == 0:
            if is_creator:
                self._epr_create_requests[request_key].pop(0)
            else:
                self._epr_recv_requests[request_key].pop(0)

    def _store_ent_info(
        self, epr_cmd_data: EprCmdData, response: T_LinkLayerResponseOK, pair_index: int
    ) -> None:
        # Store the entanglement information
        ent_info = [
            entry.value if isinstance(entry, Enum) else entry for entry in response
        ]
        ent_results_array_address = epr_cmd_data.ent_results_array_address
        self._logger.debug(
            f"Storing entanglement information for pair {pair_index} "
            f"in array at address {ent_results_array_address}"
        )
        # Start and stop of slice
        arr_start = pair_index * OK_FIELDS
        arr_stop = (pair_index + 1) * OK_FIELDS
        subroutine_id = epr_cmd_data.subroutine_id
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        if app_id not in self._app_arrays:
            raise KeyError("App ID {app_id} does not have any arrays")
        self._app_arrays[app_id][
            ent_results_array_address, arr_start:arr_stop
        ] = ent_info

    def _handle_epr_ok_k_response(
        self, epr_cmd_data: EprCmdData, response: LinkLayerOKTypeK, pair_index: int
    ) -> bool:
        # Extract qubit addresses
        subroutine_id = epr_cmd_data.subroutine_id
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        virtual_address = self._get_virtual_address_from_epr_data(
            epr_cmd_data, pair_index, app_id
        )

        # If the virtual address is currently in use, we should wait
        if self._has_virtual_address(app_id=app_id, virtual_address=virtual_address):
            self._logger.debug(
                f"Since virtual address {virtual_address} is in use, "
                "handling of epr will wait and try again."
            )
            return False

        # Update qubit mapping
        physical_address = response.logical_qubit_id
        self._logger.debug(
            f"Virtual qubit address {virtual_address} will now be mapped to "
            f"physical address {physical_address}"
        )
        self._used_physical_qubit_addresses.add(physical_address)
        self._allocate_physical_qubit(
            subroutine_id=subroutine_id,
            virtual_address=virtual_address,
            physical_address=physical_address,
        )

        return True

    def _get_virtual_address_from_epr_data(
        self, epr_cmd_data: EprCmdData, pair_index: int, app_id: int
    ) -> int:
        q_array_address = epr_cmd_data.q_array_address
        array_entry = parse_address(f"@{q_array_address}[{pair_index}]")
        assert isinstance(array_entry, ArrayEntry)
        virtual_address = self._get_array_entry(app_id=app_id, array_entry=array_entry)
        if virtual_address is None:
            raise RuntimeError("virtual address is None")
        return virtual_address

    def _handle_epr_ok_m_response(
        self, epr_cmd_data: EprCmdData, response: LinkLayerOKTypeM, pair_index: int
    ) -> bool:
        # M request are always handled
        return True

    def _handle_epr_ok_r_response(self, response: LinkLayerOKTypeR) -> bool:
        raise NotImplementedError

    def _get_qubit(self, app_id: int, virtual_address: int) -> Any:
        raise NotImplementedError

    def _get_qubit_state(self, app_id: int, virtual_address: int) -> Any:
        raise NotImplementedError

    def _get_positions(self, subroutine_id: int, addresses: List[int]) -> List[int]:
        return [
            self._get_position(subroutine_id=subroutine_id, address=address)
            for address in addresses
        ]

    def _get_position(
        self,
        subroutine_id: Optional[int] = None,
        address: int = 0,
        app_id: Optional[int] = None,
    ) -> int:
        if app_id is None:
            if subroutine_id is None:
                raise ValueError("subroutine_id and app_id cannot both be None")
            app_id = self._get_app_id(subroutine_id=subroutine_id)
        return self._get_position_in_unit_module(app_id=app_id, address=address)



"""Conversion from Python code into an NetQASM subroutines.

This module contains the `Builder` class, which is used by a Connection to transform
Python application script code into NetQASM subroutines.
"""

from __future__ import annotations

from contextlib import contextmanager
from itertools import count
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from netqasm.backend.network_stack import OK_FIELDS_K, OK_FIELDS_M
from netqasm.lang import operand
from netqasm.lang.encoding import RegisterName
from netqasm.lang.ir import (
    BranchLabel,
    BreakpointAction,
    BreakpointRole,
    GenericInstr,
    ICmd,
    ProtoSubroutine,
    T_ProtoOperand,
    flip_branch_instr,
)
from netqasm.lang.operand import Address, ArrayEntry, ArraySlice, Label, Template
from netqasm.lang.parsing.text import assemble_subroutine, parse_register
from netqasm.lang.subroutine import Subroutine
from netqasm.lang.version import NETQASM_VERSION
from netqasm.qlink_compat import BellState, EPRRole, EPRType, LinkLayerOKTypeK
from netqasm.sdk.build_epr import (
    SER_RESPONSE_KEEP_IDX_BELL_STATE,
    SER_RESPONSE_KEEP_LEN,
    EntRequestParams,
    EprKeepResult,
    EprMeasureResult,
    deserialize_epr_keep_results,
    deserialize_epr_measure_results,
    serialize_request,
)
from netqasm.sdk.build_nv import NVEprCompiler
from netqasm.sdk.build_types import (
    GenericHardwareConfig,
    HardwareConfig,
    NVHardwareConfig,
    T_BranchRoutine,
    T_CleanupRoutine,
    T_LoopRoutine,
    T_PostRoutine,
)
from netqasm.sdk.config import LogConfig
from netqasm.sdk.constraint import SdkConstraint, ValueAtMostConstraint
from netqasm.sdk.futures import Array, Future, RegFuture, T_CValue
from netqasm.sdk.memmgr import MemoryManager
from netqasm.sdk.qubit import FutureQubit, Qubit, QubitMeasureBasis
from netqasm.sdk.toolbox import get_angle_spec_from_float
from netqasm.sdk.transpile import NVSubroutineTranspiler, SubroutineTranspiler
from netqasm.typedefs import T_Cmd
from netqasm.util.log import LineTracker

if TYPE_CHECKING:
    from netqasm.sdk.connection import BaseNetQASMConnection


class LabelManager:
    """Simple manager class for providing unique branch labels."""

    def __init__(self) -> None:
        self._labels: Set[str] = set()

    def new_label(self, start_with: str = "") -> str:
        if start_with not in self._labels:
            self._labels.add(start_with)
            return start_with
        else:
            for i in count(1):
                name = f"{start_with}{i}"
                if name not in self._labels:
                    self._labels.add(name)
                    break
            return name


class SdkIfContext:
    """Context object for if statements in SDK code such as `with conn.if_eq()`."""

    def __init__(
        self,
        id: int,
        builder: Builder,
        condition: GenericInstr,
        op0: Optional[T_CValue],
        op1: Optional[T_CValue],
    ):
        self._id = id
        self._builder = builder
        self._condition = condition
        self._op0 = op0
        self._op1 = op1

    def __enter__(self):
        return self._builder.if_context_enter(context_id=self._id)

    def __exit__(self, *args, **kwargs):
        return self._builder.if_context_exit(
            context_id=self._id, condition=self._condition, op0=self._op0, op1=self._op1
        )


class SdkForEachContext:
    """Context object for foreach statements in SDK code such as `with conn.foreach()`."""

    def __init__(
        self,
        id: int,
        builder: Builder,
        array: Array,
        return_index: bool,
    ):
        self._id = id
        self._builder = builder
        self._array = array
        self._return_index = return_index

    def __enter__(self):
        return self._builder._foreach_context_enter(
            context_id=self._id, array=self._array, return_index=self._return_index
        )

    def __exit__(self, *args, **kwargs):
        return self._builder._foreach_context_exit(
            context_id=self._id, array=self._array
        )


class SdkLoopUntilContext:
    """Context object for loop_until() statements in SDK code."""

    def __init__(self, id: int, builder: Builder, max_iterations: int):
        self._id = id
        self._builder = builder
        self._exit_condition: Optional[SdkConstraint] = None
        self._cleanup_code: Optional[T_CleanupRoutine] = None
        self._loop_register: Optional[RegFuture] = None
        self._max_iterations = max_iterations

    def set_exit_condition(self, constraint: SdkConstraint) -> None:
        """Set the exit condition for this while loop."""
        self._exit_condition = constraint

    @property
    def exit_condition(self) -> Optional[SdkConstraint]:
        """Get the exit condition for this while loop."""
        return self._exit_condition

    def set_cleanup_code(self, cleanup_code: T_CleanupRoutine) -> None:
        """Set the cleanup code for this while loop."""
        self._cleanup_code = cleanup_code

    @property
    def cleanup_code(self) -> Optional[T_CleanupRoutine]:
        """Get the cleanup code for this while loop."""
        return self._cleanup_code

    def set_loop_register(self, register: RegFuture) -> None:
        """Set the register used that holds the iteration index for this while loop."""
        self._loop_register = register

    @property
    def loop_register(self) -> Optional[RegFuture]:
        """Get the register used that holds the iteration index for this while loop."""
        return self._loop_register

    @property
    def max_iterations(self) -> int:
        """Get the maximum number of iterations for this while loop."""
        return self._max_iterations


class Builder:
    """Object that transforms Python script code into `ProtoSubroutine`s.

    A Connection uses a Builder to handle statements in application script code.
    The Builder converts the statements into pseudo-NetQASM instructions that are
    assembled into a ProtoSubroutine. When the connectin flushes, the ProtoSubroutine is
    is compiled into a NetQASM subroutine.
    """

    def __init__(
        self,
        connection: BaseNetQASMConnection,
        app_id: int,
        hardware_config: Optional[HardwareConfig] = None,
        log_config: LogConfig = None,
        compiler: Optional[Type[SubroutineTranspiler]] = None,
        return_arrays: bool = True,
    ):
        """Builder constructor. Typically not used directly by the Host script.

        :param connection: Connection that this builder builds for
        :param app_id: ID of the application as given by the quantum node controller
        :param max_qubits: maximum number of qubits allowed (as registered with the
            quantum node controller)
        :param log_config: logging configuration, typically just passed as-is by the
            connection object
        :param compiler: which compiler class to use for the translation from
            ProtoSubroutine to Subroutine
        :param return_arrays: whether to add ret_arr NetQASM instructions at the end of
            each subroutine (for all arrays that are used in the subroutine). May be
            set to False if the quantum node controller does not support returning
            arrays.
        """
        self._connection = connection
        self._app_id = app_id

        self._pending_commands: List[T_Cmd] = []

        self._mem_mgr: MemoryManager = MemoryManager()

        # If False, don't return arrays even if they are used in a subroutine
        self._return_arrays: bool = return_arrays

        self._next_context_id: int = 0
        # Storing commands before an conditional statement
        self._pre_context_commands: Dict[int, List[T_Cmd]] = {}
        self._pre_context_registers: Dict[int, List[operand.Register]] = {}

        self._label_mgr = LabelManager()

        # Can be set to false for e.g. debugging, not exposed to user atm
        self._clear_app_on_exit: bool = True
        self._stop_backend_on_exit: bool = True

        if log_config is None:
            log_config = LogConfig()

        self._line_tracker: LineTracker = LineTracker(log_config=log_config)
        self._track_lines: bool = log_config.track_lines

        # Commited subroutines saved for logging/debugging
        self._committed_subroutines: List[Subroutine] = []

        self._hardware_config = hardware_config
        if self._hardware_config is None:
            self._hardware_config = GenericHardwareConfig(5)

        self._max_qubits: int = self._hardware_config.qubit_count

        # What compiler (if any) to be used
        self._compiler: Optional[Type[SubroutineTranspiler]] = compiler

        # If an NV compiler is specified but not an NV hardware config,
        # make sure an NV config is used after all.
        if compiler == NVSubroutineTranspiler:
            num_qubits = self._hardware_config.qubit_count
            self._hardware_config = NVHardwareConfig(num_qubits)

    @property
    def app_id(self) -> int:
        return self._app_id

    @app_id.setter
    def app_id(self, id: int) -> None:
        self._app_id = id

    def inactivate_qubits(self) -> None:
        self._mem_mgr.inactivate_qubits()

    def new_qubit_id(self) -> int:
        return self._mem_mgr.get_new_qubit_address()

    def alloc_array(
        self, length: int = 1, init_values: Optional[List[Optional[int]]] = None
    ) -> Array:
        address = self._mem_mgr.get_new_array_address()
        lineno = self._line_tracker.get_line()
        array = Array(
            connection=self._connection,
            length=length,
            address=address,
            init_values=init_values,
            lineno=lineno,
        )
        self._mem_mgr.add_array_to_return(array)
        return array

    def new_register(self, init_value: int = 0, return_reg: bool = True) -> RegFuture:
        reg = self._mem_mgr.get_inactive_register(activate=True)
        self.subrt_add_pending_command(
            ICmd(instruction=GenericInstr.SET, operands=[reg, init_value])
        )
        reg_future = RegFuture(connection=self._connection, reg=reg)
        if return_reg:
            self._mem_mgr.add_register_to_return(reg)
        return reg_future

    def subrt_add_pending_commands(self, commands: List[T_Cmd]) -> None:
        calling_lineno = self._line_tracker.get_line()
        for command in commands:
            if command.lineno is None:
                command.lineno = calling_lineno
            self.subrt_add_pending_command(command)

    def subrt_add_pending_command(self, command: T_Cmd) -> None:
        assert isinstance(command, ICmd) or isinstance(command, BranchLabel)
        if command.lineno is None:
            command.lineno = self._line_tracker.get_line()
        self._pending_commands.append(command)

    def subrt_pop_all_pending_commands(self) -> List[T_Cmd]:
        commands = self._pending_commands
        self._pending_commands = []
        return commands

    def subrt_pop_pending_subroutine(self) -> Optional[ProtoSubroutine]:
        # Add commands for initialising and returning arrays
        self._build_cmds_allocated_arrays()
        self._build_cmds_return_registers()
        if len(self._pending_commands) > 0:
            commands = self.subrt_pop_all_pending_commands()
            return ProtoSubroutine(
                commands=commands,
                netqasm_version=NETQASM_VERSION,
                app_id=self.app_id,
            )
        else:
            return None

    def subrt_compile_subroutine(self, pre_subroutine: ProtoSubroutine) -> Subroutine:
        """Convert a ProtoSubroutine into a Subroutine."""
        subroutine: Subroutine = assemble_subroutine(pre_subroutine)
        if self._compiler is not None:
            subroutine = self._compiler(subroutine=subroutine).transpile()
        if self._track_lines:
            self._log_subroutine(subroutine=subroutine)
        return subroutine

    def _log_subroutine(self, subroutine: Subroutine) -> None:
        self._committed_subroutines.append(subroutine)

    @property
    def committed_subroutines(self) -> List[Subroutine]:
        return self._committed_subroutines

    def _get_qubit_register(self, reg_index: int = 0) -> operand.Register:
        return operand.Register(RegisterName.Q, reg_index)

    def _alloc_epr_create_args(self, tp: EPRType, params: EntRequestParams) -> Array:
        serialized_args = serialize_request(tp, params)
        return self.alloc_array(
            length=len(serialized_args), init_values=serialized_args
        )

    def _build_cmds_wait_move_epr_to_mem(
        self, params: EntRequestParams, ent_results_array: Array, role: EPRRole
    ) -> None:

        loop_register = self._mem_mgr.get_inactive_register(activate=True)
        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)

        def post_loop(conn: BaseNetQASMConnection, loop_reg: RegFuture):
            # Wait for each pair individually
            pair = loop_register
            conn.builder._add_wait_for_ent_info_cmd(
                ent_results_array=ent_results_array,
                pair=pair,
            )

            if params.expect_phi_plus and role == EPRRole.RECV:
                # Perform Bell corrections
                bell_state = self._get_raw_bell_state(
                    ent_results_array, loop_reg, bell_state_reg
                )
                set_qubit_reg_cmd = ICmd(
                    instruction=GenericInstr.SET, operands=[qubit_reg, 0]
                )
                self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
                self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)

            # If it's the last pair, don't move it to a mem qubit
            with loop_reg.if_ne(params.number - 1):
                reg0 = self._mem_mgr.get_inactive_register(activate=True)
                reg1 = self._mem_mgr.get_inactive_register(activate=True)
                assert loop_reg.reg is not None

                # Calculate the virtual ID of the memory qubit this state should move to.
                # It is "number of pairs" - 1 - "current index".
                sub_cmd = ICmd(
                    instruction=GenericInstr.SUB,
                    operands=[reg0, params.number - 1, loop_reg.reg],
                )
                set_0_cmds = ICmd(instruction=GenericInstr.SET, operands=[reg1, 0])

                # Move the state from the communication qubit (ID = 0) to the
                # memory qubit for which we calculated the ID above.
                mov_cmd = ICmd(
                    instruction=GenericInstr.MOV,
                    operands=[reg1, reg0],
                )

                # Mark the communication qubit as free.
                free_cmd = ICmd(instruction=GenericInstr.QFREE, operands=[reg1])

                # Add the commands to the subroutine.
                commands = [sub_cmd] + [set_0_cmds] + [mov_cmd] + [free_cmd]  # type: ignore
                self.subrt_add_pending_commands(commands)  # type: ignore

                self._mem_mgr.remove_active_register(reg0)
                self._mem_mgr.remove_active_register(reg1)

        self._build_cmds_loop_body(
            post_loop, stop=params.number, loop_register=loop_register
        )
        self._mem_mgr.remove_active_register(loop_register)
        self._mem_mgr.remove_active_register(qubit_reg)
        self._mem_mgr.remove_active_register(bell_state_reg)

    def _build_cmds_post_epr(
        self,
        qubit_ids: Array,
        params: EntRequestParams,
        ent_results_array: Array,
        tp: EPRType,
        role: EPRRole,
    ) -> None:

        loop_register = self._mem_mgr.get_inactive_register(activate=True)
        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)

        def post_loop(conn: BaseNetQASMConnection, loop_reg: RegFuture):
            # Wait for each pair individually
            pair = loop_register
            conn.builder._add_wait_for_ent_info_cmd(
                ent_results_array=ent_results_array,
                pair=pair,
            )
            assert tp == EPRType.K or tp == EPRType.R

            if params.expect_phi_plus and role == EPRRole.RECV:
                bell_state = self._get_raw_bell_state(
                    ent_results_array, loop_reg, bell_state_reg
                )
                set_qubit_reg_cmd = ICmd(
                    instruction=GenericInstr.SET, operands=[qubit_reg, 0]
                )
                self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
                self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)

            q_id = qubit_ids.get_future_index(loop_register)
            q = FutureQubit(conn=conn, future_id=q_id)
            pair_future = RegFuture(self._connection, loop_register)
            assert params.post_routine is not None
            params.post_routine(self, q, pair_future)

        # TODO use loop context
        self._build_cmds_loop_body(
            post_loop, stop=params.number, loop_register=loop_register
        )
        self._mem_mgr.remove_active_register(loop_register)
        self._mem_mgr.remove_active_register(qubit_reg)
        self._mem_mgr.remove_active_register(bell_state_reg)

    def _add_wait_for_ent_info_cmd(
        self, ent_results_array: Array, pair: operand.Register
    ) -> None:
        """Wait for the correct slice of the entanglement info array for the given pair"""
        # NOTE arr_start should be pair * OK_FIELDS and
        # arr_stop should be (pair + 1) * OK_FIELDS
        arr_start = self._mem_mgr.get_inactive_register(activate=True)
        tmp = self._mem_mgr.get_inactive_register(activate=True)
        arr_stop = self._mem_mgr.get_inactive_register(activate=True)
        created_regs = [arr_start, tmp, arr_stop]

        for reg in created_regs:
            self.subrt_add_pending_command(
                ICmd(
                    instruction=GenericInstr.SET,
                    operands=[reg, 0],
                )
            )

        # Multiply pair * OK_FIELDS
        # TODO use loop context
        def add_arr_start(conn, _):
            self.subrt_add_pending_command(
                ICmd(
                    instruction=GenericInstr.ADD,
                    operands=[arr_start, arr_start, pair],
                )
            )

        self._build_cmds_loop_body(add_arr_start, stop=OK_FIELDS_K)

        # Let tmp be pair + 1
        self.subrt_add_pending_command(
            ICmd(
                instruction=GenericInstr.ADD,
                operands=[tmp, pair, 1],
            )
        )

        # Multiply (tmp = pair + 1) * OK_FIELDS
        # TODO use loop context
        def add_arr_stop(conn, _):
            self.subrt_add_pending_command(
                ICmd(
                    instruction=GenericInstr.ADD,
                    operands=[arr_stop, arr_stop, tmp],
                )
            )

        self._build_cmds_loop_body(add_arr_stop, stop=OK_FIELDS_K)

        wait_cmd = ICmd(
            instruction=GenericInstr.WAIT_ALL,
            operands=[
                ArraySlice(
                    Address(ent_results_array.address), start=arr_start, stop=arr_stop
                )
            ],
        )
        self.subrt_add_pending_command(wait_cmd)

        for reg in created_regs:
            self._mem_mgr.remove_active_register(reg)

    def _pre_epr_context(
        self,
        role: EPRRole,
        params: EntRequestParams,
    ) -> Tuple[List[T_Cmd], operand.Register, Array, FutureQubit, operand.Register]:
        self._assert_epr_args(
            number=params.number,
            post_routine=lambda: None,  # type: ignore
            sequential=params.sequential,
        )

        # Entanglement results array.
        # This will be filled in by the quantum node controller.
        ent_results_array = self._alloc_ent_results_array(
            number=params.number, tp=EPRType.K
        )

        qubit_futures = self._get_qubit_futures(
            params.number, params.sequential, ent_results_array
        )
        assert all(isinstance(q, Qubit) for q in qubit_futures)

        # NetQASM array with IDs for the generated qubits.
        virtual_qubit_ids = [q.qubit_id for q in qubit_futures]
        qubit_ids_array = self.alloc_array(init_values=virtual_qubit_ids)  # type: ignore

        # Construct and add the NetQASM instructions
        if role == EPRRole.CREATE:
            # NetQASM array for entanglement request parameters.
            create_args_array: Array = self._alloc_epr_create_args(EPRType.K, params)

            self._build_cmds_epr_create_keep(
                create_args_array, qubit_ids_array, ent_results_array, False, params
            )
        else:
            self._build_cmds_epr_recv_keep(
                qubit_ids_array, ent_results_array, False, params
            )

        pre_commands = self.subrt_pop_all_pending_commands()
        loop_register = self._mem_mgr.get_inactive_register(activate=True)
        pair = loop_register

        q_id = qubit_ids_array.get_future_index(pair)
        q = FutureQubit(conn=self._connection, future_id=q_id)

        return pre_commands, loop_register, ent_results_array, q, pair

    def _post_epr_context(
        self,
        pre_commands: List[T_Cmd],
        number: int,
        loop_register: operand.Register,
        ent_results_array: Array,
        pair: operand.Register,
    ) -> None:
        body_commands = self.subrt_pop_all_pending_commands()
        self._add_wait_for_ent_info_cmd(
            ent_results_array=ent_results_array,
            pair=pair,
        )
        wait_cmds = self.subrt_pop_all_pending_commands()
        body_commands = wait_cmds + body_commands
        self._build_cmds_loop(
            pre_commands=pre_commands,
            body_commands=body_commands,
            stop=number,
            start=0,
            step=1,
            loop_register=loop_register,
        )
        self._mem_mgr.remove_active_register(loop_register)

    def _assert_epr_args(
        self,
        number: int,
        post_routine: Optional[T_PostRoutine],
        sequential: bool,
    ) -> None:
        if sequential and number > 1:
            if post_routine is None:
                raise ValueError(
                    "When using sequential mode with more than one pair "
                    "a post_routine needs to be specified which consumes the "
                    "generated pair as they come in."
                )
        if not sequential and number > self._max_qubits:
            raise ValueError(
                f"When not using sequential mode for K type, the number of pairs {number} cannot be "
                f"greater than the maximum number of qubits specified ({self._max_qubits})."
            )

    def _check_epr_args(self, tp: EPRType, params: EntRequestParams) -> None:
        assert isinstance(tp, EPRType), "tp is not an EPRType"
        if params.sequential and params.number > 1:
            if params.post_routine is None:
                raise ValueError(
                    "When using sequential mode with more than one pair "
                    "a post_routine needs to be specified which consumes the "
                    "generated pair as they come in."
                )
        if (
            tp == EPRType.K
            and not params.sequential
            and params.number > self._max_qubits
        ):
            raise ValueError(
                f"When not using sequential mode for K type, the number of pairs "
                f"{params.number} cannot be "
                f"greater than the maximum number of qubits specified "
                f"({self._max_qubits})."
            )

    def _alloc_ent_results_array(self, number: int, tp: EPRType) -> Array:
        if tp == EPRType.K:
            ent_results_array = self.alloc_array(length=OK_FIELDS_K * number)
        elif tp == EPRType.M:
            ent_results_array = self.alloc_array(length=OK_FIELDS_M * number)
        elif tp == EPRType.R:
            # NOTE: also for R-type request we use the LinkLayerOkTypeM type
            ent_results_array = self.alloc_array(length=OK_FIELDS_M * number)
        else:
            raise ValueError(f"Unsupported Create type: {tp}")
        return ent_results_array

    def _get_qubit_futures(
        self, number: int, sequential: bool, ent_results_array: Array
    ) -> List[Qubit]:
        ent_info_slices = self._create_ent_info_k_slices(
            num_pairs=number, ent_results_array=ent_results_array
        )
        qubits = self._create_ent_qubits(
            ent_info_slices=ent_info_slices,
            sequential=sequential,
        )
        return qubits

    def _create_ent_info_k_slices(
        self, num_pairs: int, ent_results_array: Array
    ) -> List[LinkLayerOKTypeK]:
        ent_info_slices = []
        for i in range(num_pairs):
            ent_info_slice_futures: List[Future] = ent_results_array.get_future_slice(
                slice(i * OK_FIELDS_K, (i + 1) * OK_FIELDS_K)
            )
            ent_info_slice = LinkLayerOKTypeK(*ent_info_slice_futures)
            ent_info_slices.append(ent_info_slice)
        return ent_info_slices

    def _create_ent_qubits(
        self, ent_info_slices: List[LinkLayerOKTypeK], sequential: bool
    ) -> List[Qubit]:
        qubits: List[Qubit] = []
        num_pairs = len(ent_info_slices)

        if isinstance(self._hardware_config, NVHardwareConfig):
            # NV: only ID 0 can be used for entanglement
            self._build_cmds_free_up_qubit_location(0)
            if sequential:
                # all qubits get ID 0
                for _, ent_info_slice in enumerate(ent_info_slices):
                    q = Qubit(
                        self._connection,
                        add_new_command=False,
                        ent_info=ent_info_slice,
                        virtual_address=0,
                    )
                    qubits.append(q)
            else:  # not sequential
                # EXPLICIT MOVE (i.e. application moves states from comm to mem qubits)

                # Create the Qubit object that is returned to the SDK.
                for i, ent_info_slice in enumerate(ent_info_slices):
                    # Allocate and initialize the memory qubit which the entangled
                    # state should finally end up in.

                    final_id = num_pairs - 1 - i

                    # If the final qubit is the comm qubit (ID 0), don't allocate it,
                    # since it will automatically be allocated as part of
                    # entanglement generation.
                    add_new_command = True
                    if final_id == 0:
                        add_new_command = False
                    else:
                        # Make sure the memory qubit is free so we can allocate it.
                        assert not self._mem_mgr.is_qubit_id_used(final_id)

                    q = Qubit(
                        self._connection,
                        add_new_command=add_new_command,
                        ent_info=ent_info_slice,
                        virtual_address=final_id,
                    )
                    qubits.append(q)
        else:  # generic hardware
            virt_id: Optional[int]
            if sequential:
                # use one and the same virtual ID for all qubits
                virt_id = self._mem_mgr.get_new_qubit_address()
            else:
                virt_id = None  # let Qubit constructor choose unused ID
            for i, ent_info_slice in enumerate(ent_info_slices):
                q = Qubit(
                    self._connection,
                    add_new_command=False,
                    ent_info=ent_info_slice,
                    virtual_address=virt_id,
                )
                qubits.append(q)
        return qubits

    def _reset(self) -> None:
        # if len(self._active_registers) > 0:
        #     raise RuntimeError("Should not have active registers left when flushing")
        self._mem_mgr.reset()
        self._pre_context_commands = {}

    def _get_condition_operand(
        self, value: T_CValue
    ) -> Tuple[List[ICmd], T_ProtoOperand]:
        if isinstance(value, Future):
            # Register for checking branching based on condition
            reg = self._mem_mgr.get_inactive_register(activate=True)
            # Load values
            address_entry = value.get_address_entry()
            load = ICmd(instruction=GenericInstr.LOAD, operands=[reg, address_entry])
            return [load], reg
        elif isinstance(value, RegFuture):
            assert value.reg is not None
            return [], value.reg
        elif isinstance(value, int):
            return [], value
        else:
            raise TypeError(f"Cannot do conditional statement with type {type(value)}")

    def _get_branch_commands_single_operand(
        self,
        branch_instruction: GenericInstr,
        op: T_CValue,
    ) -> Tuple[List[ICmd], List[BranchLabel]]:
        # Exit label
        exit_label = self._label_mgr.new_label(start_with="IF_EXIT")
        if_start: List[ICmd] = []

        using_new_temp_reg = False

        cmds, cond_operand = self._get_condition_operand(op)
        if_start.extend(cmds)

        branch = ICmd(
            instruction=branch_instruction,
            operands=[cond_operand, Label(exit_label)],
        )
        if_start.append(branch)

        if using_new_temp_reg:
            assert isinstance(cond_operand, operand.Register)
            self._mem_mgr.remove_active_register(cond_operand)

        exit = BranchLabel(exit_label)
        if_end = [exit]

        return if_start, if_end

    def _get_branch_commands(
        self,
        branch_instruction: GenericInstr,
        op0: T_CValue,
        op1: T_CValue,
    ) -> Tuple[List[ICmd], List[BranchLabel]]:
        # Exit label
        exit_label = self._label_mgr.new_label(start_with="IF_EXIT")
        cond_operands: List[T_ProtoOperand] = []
        if_start: List[ICmd] = []

        temp_regs_to_remove: List[operand.Register] = []

        for x in [op0, op1]:
            cmds, cond_operand = self._get_condition_operand(x)
            if_start.extend(cmds)
            cond_operands.append(cond_operand)

            if isinstance(x, Future):
                assert isinstance(cond_operand, operand.Register)
                temp_regs_to_remove.append(cond_operand)

        branch = ICmd(
            instruction=branch_instruction,
            operands=[cond_operands[0], cond_operands[1], Label(exit_label)],
        )
        if_start.append(branch)

        # Inactivate the temporary registers
        for reg in temp_regs_to_remove:
            self._mem_mgr.remove_active_register(reg)

        exit = BranchLabel(exit_label)
        if_end = [exit]

        return if_start, if_end

    @contextmanager
    def _activate_register(self, register: operand.Register) -> Iterator[None]:
        try:
            self._mem_mgr.add_active_register(register)
            yield
        except Exception as err:
            raise err
        finally:
            self._mem_mgr.remove_active_register(register)

    def _loop_get_register(
        self,
        register: Optional[Union[operand.Register, str]],
        activate: bool = False,
    ) -> operand.Register:
        if register is None:
            return self._mem_mgr.get_inactive_register(activate=activate)

        assert isinstance(register, operand.Register) or isinstance(
            register, str
        ), f"not a valid loop_register with type {type(register)}"
        if isinstance(register, str):
            loop_register = parse_register(register)
        else:
            loop_register = register

        # if self._mem_mgr.is_register_active(loop_register):
        #     raise ValueError("Register used for looping should not already be active")
        return loop_register

    def _loop_get_entry_commands(
        self,
        start: int,
        stop: int,
        entry_label: str,
        exit_label: str,
        loop_register: operand.Register,
    ) -> List[T_Cmd]:
        return [
            ICmd(instruction=GenericInstr.SET, operands=[loop_register, start]),
            BranchLabel(entry_label),
            ICmd(
                instruction=GenericInstr.BEQ,
                operands=[loop_register, stop, Label(exit_label)],
            ),
        ]

    def _loop_get_exit_commands(
        self,
        step: int,
        entry_label: str,
        exit_label: str,
        loop_register: operand.Register,
    ) -> List[T_Cmd]:
        return [
            ICmd(
                instruction=GenericInstr.ADD,
                operands=[loop_register, loop_register, step],
            ),
            ICmd(instruction=GenericInstr.JMP, operands=[Label(entry_label)]),
            BranchLabel(exit_label),
        ]

    def _loop_until_get_entry_commands(
        self,
        entry_label: str,
        exit_label: str,
        stop: int,
        loop_register: operand.Register,
    ) -> List[T_Cmd]:
        return [
            ICmd(instruction=GenericInstr.SET, operands=[loop_register, 0]),
            BranchLabel(entry_label),
            ICmd(
                instruction=GenericInstr.BEQ,
                operands=[loop_register, stop, Label(exit_label)],
            ),
        ]

    def _loop_until_get_break_commands(
        self,
        context: SdkLoopUntilContext,
        exit_label: str,
    ) -> List[T_Cmd]:
        commands: List[ICmd] = []
        condition = context.exit_condition

        if isinstance(condition, ValueAtMostConstraint):
            if_start: List[ICmd] = []

            temp_regs_to_remove: List[operand.Register] = []

            cmds, cond_operand = self._get_condition_operand(condition.future)
            if_start.extend(cmds)

            if isinstance(condition.future, Future):
                assert isinstance(cond_operand, operand.Register)
                temp_regs_to_remove.append(cond_operand)

            branch = ICmd(
                instruction=GenericInstr.BLT,
                operands=[cond_operand, condition.value, Label(exit_label)],
            )
            if_start.append(branch)
            commands = if_start
        else:
            assert False, "not supported"
        return commands  # type: ignore

    def _loop_until_get_exit_commands(
        self, entry_label: str, exit_label: str, loop_register: operand.Register
    ) -> List[T_Cmd]:
        return [
            ICmd(
                instruction=GenericInstr.ADD,
                operands=[loop_register, loop_register, 1],
            ),
            ICmd(instruction=GenericInstr.JMP, operands=[Label(entry_label)]),
            BranchLabel(exit_label),
        ]

    def if_context_enter(self, context_id: int) -> None:
        pre_commands = self.subrt_pop_all_pending_commands()
        self._pre_context_commands[context_id] = pre_commands

    def if_context_exit(
        self,
        context_id: int,
        condition: GenericInstr,
        op0: T_CValue,
        op1: Optional[T_CValue],
    ) -> None:
        # pop commands that were added while evaluting the context body
        body_commands = self.subrt_pop_all_pending_commands()

        # get all commands that were pending before entering this context
        pre_context_commands = self._pre_context_commands.pop(context_id, None)
        if pre_context_commands is None:
            raise RuntimeError("Something went wrong, no pre_context_commands")

        self._build_cmds_condition(
            pre_commands=pre_context_commands,
            body_commands=body_commands,
            condition=condition,
            op0=op0,
            op1=op1,
        )

    def _foreach_context_enter(
        self, context_id: int, array: Array, return_index: bool
    ) -> Union[Tuple[operand.Register, Future], Future]:
        pre_commands = self.subrt_pop_all_pending_commands()
        loop_register = self._mem_mgr.get_inactive_register(activate=True)

        # NOTE (BUG): the below assignment is NOT consistent with the type of _pre_context_commands
        # It works (maybe?) because the values are pushed only temporarily
        self._pre_context_commands[context_id] = pre_commands, loop_register  # type: ignore
        if return_index:
            return loop_register, array.get_future_index(loop_register)
        else:
            return array.get_future_index(loop_register)

    def _foreach_context_exit(self, context_id: int, array: Array) -> None:
        body_commands = self.subrt_pop_all_pending_commands()
        pre_context_commands: Tuple[List[T_Cmd], operand.Register] = self._pre_context_commands.pop(  # type: ignore
            context_id, None  # type: ignore
        )
        if pre_context_commands is None:
            raise RuntimeError("Something went wrong, no pre_context_commands")

        # NOTE (BUG): see NOTE (BUG) in _foreach_context_enter
        pre_commands, loop_register = pre_context_commands  # type: ignore
        self._build_cmds_loop(
            pre_commands=pre_commands,
            body_commands=body_commands,
            stop=len(array),
            start=0,
            step=1,
            loop_register=loop_register,
        )
        self._mem_mgr.remove_active_register(loop_register)

    def _loop_until_context_enter(self, context_id: int) -> operand.Register:
        pre_commands = self.subrt_pop_all_pending_commands()
        loop_register = self._mem_mgr.get_inactive_register(activate=True)

        self._pre_context_commands[context_id] = pre_commands
        self._pre_context_registers[context_id] = [loop_register]

        return loop_register

    def _loop_until_context_exit(
        self, context_id: int, context: SdkLoopUntilContext
    ) -> None:
        body_commands = self.subrt_pop_all_pending_commands()
        pre_commands = self._pre_context_commands.pop(context_id, None)
        if pre_commands is None:
            raise RuntimeError("Something went wrong, no pre_commands")
        loop_registers = self._pre_context_registers.pop(context_id, None)
        if loop_registers is None:
            raise RuntimeError("Something went wrong, no loop_registers for context")
        loop_register = loop_registers[0]

        self._build_cmds_loop_until(
            pre_commands=pre_commands,
            body_commands=body_commands,
            context=context,
            loop_register=loop_register,
        )

    def _build_cmds_breakpoint(
        self, action: BreakpointAction, role: BreakpointRole = BreakpointRole.CREATE
    ) -> None:
        self.subrt_add_pending_command(
            ICmd(
                instruction=GenericInstr.BREAKPOINT, operands=[action.value, role.value]
            )
        )

    def _build_cmds_single_qubit_rotation(
        self,
        instruction: GenericInstr,
        virtual_qubit_id: int,
        n: Union[int, Template] = 0,
        d: int = 0,
        angle: Optional[float] = None,
    ) -> None:
        if angle is not None:
            nds = get_angle_spec_from_float(angle=angle)
            for n, d in nds:
                self._build_cmds_single_qubit_rotation(
                    instruction=instruction,
                    virtual_qubit_id=virtual_qubit_id,
                    n=n,
                    d=d,
                )
            return
        if not (isinstance(n, int) and isinstance(d, int) and n >= 0 and d >= 0):
            # We only allow this if n is a Template
            if not isinstance(n, Template):
                raise ValueError(
                    f"{n} * pi / 2 ^ {d} is not a valid angle specification"
                )
        register = self._get_qubit_register()
        self._build_cmds_set_register_value(register, virtual_qubit_id)
        rot_command = ICmd(
            instruction=instruction,
            operands=[register, n, d],
        )
        self.subrt_add_pending_command(rot_command)

    def _build_cmds_single_qubit(self, instr: GenericInstr, qubit_id: int) -> None:
        register = self._get_qubit_register()
        self._build_cmds_set_register_value(register, qubit_id)
        # Construct the qubit command
        qubit_command = ICmd(
            instruction=instr,
            operands=[register],
        )
        self.subrt_add_pending_command(qubit_command)

    def _build_cmds_two_qubit(
        self, instr: GenericInstr, control_qubit_id: int, target_qubit_id: int
    ) -> None:
        register1 = self._get_qubit_register(0)
        self._build_cmds_set_register_value(register1, control_qubit_id)
        register2 = self._get_qubit_register(1)
        self._build_cmds_set_register_value(register2, target_qubit_id)
        qubit_command = ICmd(
            instruction=instr,
            operands=[register1, register2],
        )
        self.subrt_add_pending_command(qubit_command)

    def _build_cmds_move_qubit(self, source: int, target: int) -> None:
        # Moves a qubit from one position to another (assumes that target is free)
        assert target not in [q.qubit_id for q in self._mem_mgr.get_active_qubits()]
        self._build_cmds_new_qubit(target)
        self._build_cmds_two_qubit(GenericInstr.MOV, source, target)
        self._build_cmds_qfree(source)

    def _build_cmds_measure(
        self,
        qubit_id: int,
        future: Union[Future, RegFuture],
        inplace: bool,
        basis: QubitMeasureBasis = QubitMeasureBasis.Z,
        rotations: Optional[Tuple[int, int, int]] = None,
    ) -> None:
        if isinstance(self._hardware_config, NVHardwareConfig):
            # If compiling for NV, only virtual ID 0 can be used to measure a qubit.
            # So, if this qubit is already in use, we need to move it away first.
            if not isinstance(qubit_id, Future):
                if qubit_id != 0:
                    self._build_cmds_free_up_qubit_location(virtual_address=0)
        outcome_reg = self._mem_mgr.get_new_meas_outcome_register()
        qubit_reg = self._get_qubit_register()
        self._build_cmds_set_register_value(qubit_reg, qubit_id)

        # use denominator 4 since we always treat angles as multiples of pi/(2^4)
        denom = 4

        if rotations is not None:
            x1, y, x2 = rotations
            meas_command = ICmd(
                instruction=GenericInstr.MEAS_BASIS,
                operands=[qubit_reg, outcome_reg, x1, y, x2, denom],
            )
        elif basis == QubitMeasureBasis.X:
            meas_command = ICmd(
                instruction=GenericInstr.MEAS_BASIS,
                operands=[qubit_reg, outcome_reg, 0, 24, 0, denom],  # -pi/2 Y rotation
            )
        elif basis == QubitMeasureBasis.Y:
            meas_command = ICmd(
                instruction=GenericInstr.MEAS_BASIS,
                operands=[qubit_reg, outcome_reg, 8, 0, 0, denom],  # pi/2 X rotation
            )
        else:
            meas_command = ICmd(
                instruction=GenericInstr.MEAS,
                operands=[qubit_reg, outcome_reg],
            )

        if not inplace:
            free_commands = [
                ICmd(
                    instruction=GenericInstr.QFREE,
                    operands=[qubit_reg],
                )
            ]
        else:
            free_commands = []
        if future is not None:
            if isinstance(future, Future):
                outcome_commands = future._get_store_commands(outcome_reg)
                self._mem_mgr.meas_register_set_unused(outcome_reg)
            elif isinstance(future, RegFuture):
                future.reg = outcome_reg
                self._mem_mgr.add_register_to_return(outcome_reg)
                outcome_commands = []
            else:
                outcome_commands = []
        commands = [meas_command] + free_commands + outcome_commands  # type: ignore
        self.subrt_add_pending_commands(commands)  # type: ignore

    def _build_cmds_new_qubit(self, qubit_id: int) -> None:
        qubit_reg = self._get_qubit_register()
        self._build_cmds_set_register_value(qubit_reg, qubit_id)
        qalloc_command = ICmd(
            instruction=GenericInstr.QALLOC,
            operands=[qubit_reg],
        )
        init_command = ICmd(
            instruction=GenericInstr.INIT,
            operands=[qubit_reg],
        )
        commands = [qalloc_command, init_command]
        self.subrt_add_pending_commands(commands)  # type: ignore

    def _build_cmds_init_qubit(self, qubit_id: int) -> None:
        qubit_reg = self._get_qubit_register()
        self._build_cmds_set_register_value(qubit_reg, qubit_id)
        init_command = ICmd(
            instruction=GenericInstr.INIT,
            operands=[qubit_reg],
        )
        self.subrt_add_pending_command(init_command)

    def _build_cmds_qfree(self, qubit_id: int) -> None:
        qubit_reg = self._get_qubit_register()
        self._build_cmds_set_register_value(qubit_reg, qubit_id)
        qfree_command = ICmd(
            instruction=GenericInstr.QFREE,
            operands=[qubit_reg],
        )
        self.subrt_add_pending_command(qfree_command)

    def _build_cmds_allocated_arrays(self) -> None:
        current_commands = self.subrt_pop_all_pending_commands()

        for array in self._mem_mgr.get_arrays_to_return():
            self._build_cmds_init_array(array)
        self.subrt_add_pending_commands(current_commands)

        if self._return_arrays:
            for array in self._mem_mgr.get_arrays_to_return():
                self._build_cmds_return_array(array)

    def _build_cmds_init_array(self, array: Array) -> None:
        commands: List[T_Cmd] = []

        array_cmd = ICmd(
            instruction=GenericInstr.ARRAY,
            operands=[len(array), Address(array.address)],
            lineno=array.lineno,
        )
        commands.append(array_cmd)

        init_vals = array._init_values
        if init_vals is not None:
            length = len(init_vals)
            if (
                length > 1
                and init_vals[0] is not None
                and init_vals.count(init_vals[0]) == length
            ):
                # Ad-hoc optimization: if all values are the same, put the initialization commands in a loop
                loop_register = self._mem_mgr.get_inactive_register()

                store_cmd = ICmd(
                    instruction=GenericInstr.STORE,
                    operands=[
                        init_vals[0],
                        ArrayEntry(Address(array.address), loop_register),
                    ],
                    lineno=array.lineno,
                )

                def init_array_elt(conn, _):
                    conn._builder.subrt_add_pending_command(store_cmd)

                self._build_cmds_loop_body(
                    init_array_elt, stop=length, loop_register=loop_register
                )
                commands.extend(self.subrt_pop_all_pending_commands())
            else:
                for i, value in enumerate(init_vals):
                    if value is None:
                        continue
                    else:
                        store_cmd = ICmd(
                            instruction=GenericInstr.STORE,
                            operands=[value, ArrayEntry(Address(array.address), i)],
                            lineno=array.lineno,
                        )
                        commands.append(store_cmd)
        self.subrt_add_pending_commands(commands)

    def _build_cmds_set_register_value(
        self, register: operand.Register, value: Union[Future, int]
    ) -> None:
        if isinstance(value, Future):
            set_reg_cmds = value.get_load_commands(register)
        elif isinstance(value, int):
            set_reg_cmds = [
                ICmd(instruction=GenericInstr.SET, operands=[register, value])
            ]
        self.subrt_add_pending_commands(set_reg_cmds)

    def _build_cmds_return_array(self, array: Array) -> None:
        self.subrt_add_pending_command(
            ICmd(
                instruction=GenericInstr.RET_ARR,
                operands=[Address(array.address)],
                lineno=array.lineno,
            )
        )

    def _build_cmds_return_registers(self) -> None:
        ret_reg_instrs: List[T_Cmd] = []
        for reg in self._mem_mgr.get_registers_to_return():
            ret_reg_instrs.append(
                ICmd(instruction=GenericInstr.RET_REG, operands=[reg])
            )
        self.subrt_add_pending_commands(commands=ret_reg_instrs)

    def _build_cmds_free_up_qubit_location(self, virtual_address: int) -> None:
        if isinstance(self._hardware_config, NVHardwareConfig):
            for q in self._mem_mgr.get_active_qubits():
                # Find a free qubit
                if q.qubit_id == virtual_address:
                    # Virtual address is already used. Move it to a new virtual address.
                    # NOTE: this assumes that the new virtual address is *not* currently used.
                    new_virtual_address = self._mem_mgr.get_new_qubit_address()

                    # Check if the existing qubit was just initialized. If so, we can
                    # simply change the address used for initialization instead of
                    # adding move instructions. I.e. instead of
                    # "init old address" + "move to new address", we just have
                    # "init new address"
                    pending_commands = self.subrt_pop_all_pending_commands()
                    # Check last 3 commands to see if the qubit was just initialized.
                    # ('type: ignore' since mypy isn't smart enough)
                    if len(pending_commands) >= 3 and (
                        all(isinstance(cmd, ICmd) for cmd in pending_commands[-3:])
                        and pending_commands[-3].instruction == GenericInstr.SET  # type: ignore
                        and pending_commands[-2].instruction == GenericInstr.QALLOC  # type: ignore
                        and pending_commands[-1].instruction == GenericInstr.INIT  # type: ignore
                    ):
                        # Update the SET command with the new address.
                        pending_commands[-3].operands[1] = new_virtual_address  # type: ignore
                        self.subrt_add_pending_commands(pending_commands)
                    else:
                        self.subrt_add_pending_commands(pending_commands)
                        # Move the existing qubit from the old to the new address.
                        self._build_cmds_move_qubit(
                            source=virtual_address, target=new_virtual_address
                        )
                    # From now on, the original qubit should be referred to with the new virtual address.
                    q.qubit_id = new_virtual_address

    def _build_cmds_undefine_array(self, array: Array) -> None:
        index_reg = self._mem_mgr.get_inactive_register()
        undef_cmd = ICmd(
            instruction=GenericInstr.UNDEF,
            operands=[ArrayEntry(Address(array.address), index_reg)],
        )

        def undef_result_element(conn: BaseNetQASMConnection, _: RegFuture):
            conn.builder.subrt_add_pending_command(undef_cmd)

        self._build_cmds_loop_body(
            undef_result_element,
            stop=len(array),
            loop_register=index_reg,
        )

    def _build_cmds_epr_keep_corrections_single_pair(
        self, bell_state: RegFuture, qubit_reg: operand.Register
    ) -> None:
        with bell_state.if_eq(BellState.PHI_MINUS.value):  # Phi- -> apply Z-gate
            correction_cmds = [
                ICmd(instruction=GenericInstr.ROT_Z, operands=[qubit_reg, 16, 4])
            ]
            self.subrt_add_pending_commands(correction_cmds)  # type: ignore
        with bell_state.if_eq(BellState.PSI_PLUS.value):  # Psi+ -> apply X-gate
            correction_cmds = [
                ICmd(instruction=GenericInstr.ROT_X, operands=[qubit_reg, 16, 4])
            ]
            self.subrt_add_pending_commands(correction_cmds)  # type: ignore
        with bell_state.if_eq(
            BellState.PSI_MINUS.value
        ):  # Psi- -> apply X-gate and Z-gate
            correction_cmds = [
                ICmd(instruction=GenericInstr.ROT_X, operands=[qubit_reg, 16, 4]),
                ICmd(instruction=GenericInstr.ROT_Z, operands=[qubit_reg, 16, 4]),
            ]
            self.subrt_add_pending_commands(correction_cmds)  # type: ignore

    def _build_cmds_epr_keep_corrections(
        self, qubit_ids_array: Array, ent_results_array: Array, params: EntRequestParams
    ) -> None:
        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)

        loop_register = self._mem_mgr.get_inactive_register()

        def loop(conn: BaseNetQASMConnection, index: RegFuture):
            qubit_reg_cmds = qubit_ids_array.get_future_index(index).get_load_commands(
                qubit_reg
            )
            self.subrt_add_pending_commands(qubit_reg_cmds)
            bell_state = self._get_raw_bell_state(
                ent_results_array, index, bell_state_reg
            )
            set_qubit_reg_cmd = ICmd(
                instruction=GenericInstr.SET, operands=[qubit_reg, 0]
            )
            self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
            self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)

        self._build_cmds_loop_body(
            loop, stop=params.number, loop_register=loop_register
        )

        self._mem_mgr.remove_active_register(qubit_reg)
        self._mem_mgr.remove_active_register(bell_state_reg)

    def _get_raw_bell_state(
        self, results_array: Array, index: RegFuture, target_reg: operand.Register
    ) -> RegFuture:
        # calculate index in array

        index_reg = self._mem_mgr.get_inactive_register(activate=True)
        # set index_reg to index (in single element) of raw_bell_state
        self.subrt_add_pending_command(
            ICmd(
                GenericInstr.SET, operands=[index_reg, SER_RESPONSE_KEEP_IDX_BELL_STATE]
            )
        )

        # add index*elt-size to it to arrive in correct element
        def add_loop(_1, _2):
            self.subrt_add_pending_command(
                ICmd(
                    GenericInstr.ADD,
                    operands=[index_reg, index_reg, SER_RESPONSE_KEEP_LEN],
                )
            )

        self._build_cmds_loop_body(add_loop, stop=index.reg)  # type: ignore

        bell_state_reg_cmds = results_array.get_future_index(
            index_reg
        ).get_load_commands(target_reg)
        self.subrt_add_pending_commands(bell_state_reg_cmds)

        self._mem_mgr.remove_active_register(index_reg)
        return RegFuture(self._connection, target_reg)

    def _build_cmds_epr_create_keep(
        self,
        create_args_array: Array,
        qubit_ids_array: Array,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            qubit_ids_array.address,
            create_args_array.address,
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.CREATE_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

    def _build_cmds_epr_recv_keep(
        self,
        qubit_ids_array: Array,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            qubit_ids_array.address,
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.RECV_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

        if wait_all and params.expect_phi_plus:
            self._build_cmds_epr_keep_corrections(
                qubit_ids_array, ent_results_array, params
            )

    def _build_cmds_epr_create_measure(
        self,
        create_args_array: Array,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            # NOTE since the qubit IDs array won't be used just set it to some
            # constant register for now
            operand.Register(RegisterName.C, 0),
            create_args_array.address,
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.CREATE_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

    def _build_cmds_epr_recv_measure(
        self,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            # NOTE since the qubit IDs array won't be used just set it to some
            # constant register for now
            operand.Register(RegisterName.C, 0),
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.RECV_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

    def _build_cmds_epr_create_rsp(
        self,
        create_args_array: Array,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            # NOTE since the qubit IDs array won't be used just set it to some
            # constant register for now
            operand.Register(RegisterName.C, 0),
            create_args_array.address,
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.CREATE_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

    def _build_cmds_epr_recv_rsp(
        self,
        qubit_ids_array: Array,
        ent_results_array: Array,
        wait_all: bool,
        params: EntRequestParams,
        **kwargs,
    ) -> None:
        epr_cmd_operands = [
            qubit_ids_array.address,
            ent_results_array.address,
        ]

        # epr command
        epr_cmd = ICmd(
            instruction=GenericInstr.RECV_EPR,
            args=[params.remote_node_id, params.epr_socket_id],
            operands=epr_cmd_operands,  # type: ignore
        )
        self.subrt_add_pending_command(epr_cmd)

        # wait
        arr_slice = ArraySlice(
            ent_results_array.address, start=0, stop=len(ent_results_array)  # type: ignore
        )
        if wait_all:
            wait_cmds = [ICmd(instruction=GenericInstr.WAIT_ALL, operands=[arr_slice])]
        else:
            wait_cmds = []

        self.subrt_add_pending_commands(wait_cmds)  # type: ignore

        if wait_all and params.expect_phi_plus:
            self._build_cmds_epr_keep_corrections(
                qubit_ids_array, ent_results_array, params
            )

    def _build_cmds_loop_body(
        self,
        body: T_LoopRoutine,
        stop: int,
        start: int = 0,
        step: int = 1,
        loop_register: Optional[Union[operand.Register, str]] = None,
    ) -> None:
        """An effective loop-statement where body is a function executed, a number of times specified
        by `start`, `stop` and `step`.
        """
        loop_register = self._loop_get_register(loop_register)
        pre_commands = self.subrt_pop_all_pending_commands()

        loop_register_already_activated: bool = self._mem_mgr.is_register_active(
            loop_register
        )

        if not loop_register_already_activated:
            self._mem_mgr.add_active_register(loop_register)
        # evaluate body (will add pending commands)
        body(
            self._connection,
            RegFuture(connection=self._connection, reg=loop_register),
        )
        body_commands = self.subrt_pop_all_pending_commands()

        self._build_cmds_loop(
            pre_commands=pre_commands,
            body_commands=body_commands,
            stop=stop,
            start=start,
            step=step,
            loop_register=loop_register,
        )
        if not loop_register_already_activated:
            self._mem_mgr.remove_active_register(loop_register)

    def _build_cmds_loop(
        self,
        pre_commands: List[T_Cmd],
        body_commands: List[T_Cmd],
        stop: int,
        start: int,
        step: int,
        loop_register: operand.Register,
    ) -> None:
        if len(body_commands) == 0:
            self.subrt_add_pending_commands(commands=pre_commands)
            return

        entry_label = self._label_mgr.new_label(start_with="LOOP")
        exit_label = self._label_mgr.new_label(start_with="LOOP_EXIT")

        loop_start = self._loop_get_entry_commands(
            start=start,
            stop=stop,
            entry_label=entry_label,
            exit_label=exit_label,
            loop_register=loop_register,
        )
        loop_end = self._loop_get_exit_commands(
            step=step,
            entry_label=entry_label,
            exit_label=exit_label,
            loop_register=loop_register,
        )

        commands = pre_commands + loop_start + body_commands + loop_end

        self.subrt_add_pending_commands(commands=commands)

    def _build_cmds_loop_until(
        self,
        pre_commands: List[T_Cmd],
        body_commands: List[T_Cmd],
        context: SdkLoopUntilContext,
        loop_register: operand.Register,
    ) -> None:
        if len(body_commands) == 0:
            self.subrt_add_pending_commands(commands=pre_commands)
            return

        entry_label = self._label_mgr.new_label(start_with="WHILE")
        exit_label = self._label_mgr.new_label(start_with="WHILE_EXIT")

        loop_until_start = self._loop_until_get_entry_commands(
            entry_label=entry_label,
            exit_label=exit_label,
            stop=context.max_iterations,
            loop_register=loop_register,
        )
        loop_until_break = self._loop_until_get_break_commands(
            context=context, exit_label=exit_label
        )

        cleanup_body = context.cleanup_code
        if cleanup_body is not None:
            cleanup_body(self._connection)
            cleanup_commands = self.subrt_pop_all_pending_commands()
        else:
            cleanup_commands = []

        loop_until_end = self._loop_until_get_exit_commands(
            entry_label=entry_label, exit_label=exit_label, loop_register=loop_register
        )

        commands = (
            pre_commands
            + loop_until_start
            + body_commands
            + loop_until_break
            + cleanup_commands
            + loop_until_end
        )

        self.subrt_add_pending_commands(commands=commands)

    def _build_cmds_if_stmt(
        self,
        condition: GenericInstr,
        op0: T_CValue,
        op1: Optional[T_CValue],
        body: T_BranchRoutine,
    ) -> None:
        """Used to build effective if-statements"""
        current_commands = self.subrt_pop_all_pending_commands()

        # evaluate body (will add pending commands)
        body(self._connection)

        # get those commands
        body_commands = self.subrt_pop_all_pending_commands()

        # combine existing commands with body commands and branch instructions
        self._build_cmds_condition(
            pre_commands=current_commands,
            body_commands=body_commands,
            condition=condition,
            op0=op0,
            op1=op1,
        )

    def _build_cmds_condition(
        self,
        pre_commands: List[T_Cmd],
        body_commands: List[T_Cmd],
        condition: GenericInstr,
        op0: T_CValue,
        op1: Optional[T_CValue],
    ) -> None:
        if len(body_commands) == 0:
            self.subrt_add_pending_commands(commands=pre_commands)
            return
        negated_predicate = flip_branch_instr(condition)
        # Construct a list of all commands to see what branch labels are already used
        all_commands = pre_commands + body_commands
        # We also need to check any existing other pre context commands if they are nested
        for pre_context_cmds in self._pre_context_commands.values():
            all_commands.extend(pre_context_cmds)

        if negated_predicate in [GenericInstr.BEZ, GenericInstr.BNZ]:
            if_start, if_end = self._get_branch_commands_single_operand(
                branch_instruction=negated_predicate, op=op0
            )
        else:
            assert op1 is not None
            if_start, if_end = self._get_branch_commands(
                branch_instruction=negated_predicate,
                op0=op0,
                op1=op1,
            )
        commands: List[T_Cmd] = pre_commands + if_start + body_commands + if_end  # type: ignore

        self.subrt_add_pending_commands(commands=commands)

    def sdk_epr_keep(
        self,
        role: EPRRole,
        params: EntRequestParams,
        reset_results_array: bool = False,
    ) -> Tuple[List[Qubit], Array]:
        """Build commands for an EPR keep operation and return the result futures."""
        self._check_epr_args(tp=EPRType.K, params=params)

        # Setup NetQASM arrays and SDK handles.

        # NetQASM array for entanglement results.
        # This will be filled in by the quantum node controller.
        ent_results_array: Array = self._alloc_ent_results_array(
            number=params.number, tp=EPRType.K
        )

        # SDK handles to result values (Qubit objects).
        qubit_futures: List[Qubit] = self._get_qubit_futures(
            params.number, params.sequential, ent_results_array
        )
        assert all(isinstance(q, Qubit) for q in qubit_futures)

        # NetQASM array with IDs for the generated qubits.
        if (
            self._hardware_config is not None
            and self._hardware_config.comm_qubit_count == 1
        ):
            # If there is only one comm qubit, only ID 0 can be used to receive the
            # EPR qubit. The Builder will however insert instructions to move this
            # qubit to one of the memory qubits. The corresponding QubitFuture object
            # still has the ID of this memory qubit (and not ID 0)!
            virtual_qubit_ids = [0 for _ in qubit_futures]
        else:
            virtual_qubit_ids = [q.qubit_id for q in qubit_futures]
        qubit_ids_array: Array = self.alloc_array(init_values=virtual_qubit_ids)  # type: ignore

        single_comm_qubit: bool = (
            self._hardware_config is not None
            and self._hardware_config.comm_qubit_count == 1
        )

        # If there is a post routine, handle pairs one by one.
        # If there is only one comm qubit, handle pairs one by one.
        if params.post_routine is not None or single_comm_qubit:
            wait_all = False
        else:
            wait_all = True

        if reset_results_array:
            self._build_cmds_undefine_array(ent_results_array)

        # Construct and add the NetQASM instructions
        if role == EPRRole.CREATE:
            # NetQASM array for entanglement request parameters.
            create_args_array: Array = self._alloc_epr_create_args(EPRType.K, params)

            self._build_cmds_epr_create_keep(
                create_args_array,
                qubit_ids_array,
                ent_results_array,
                wait_all,
                params,
            )
        else:
            self._build_cmds_epr_recv_keep(
                qubit_ids_array, ent_results_array, wait_all, params
            )

        if params.post_routine is None and single_comm_qubit:
            self._build_cmds_wait_move_epr_to_mem(
                params=params, ent_results_array=ent_results_array, role=role
            )

        # Construct and add NetQASM instructions for post routine
        if params.post_routine:
            self._build_cmds_post_epr(
                qubit_ids_array, params, ent_results_array, EPRType.K, role
            )

        return qubit_futures, ent_results_array

    def sdk_epr_measure(
        self,
        role: EPRRole,
        params: EntRequestParams,
    ) -> List[EprMeasureResult]:
        """Build commands for an EPR measure operation and return the result futures."""
        self._check_epr_args(tp=EPRType.M, params=params)

        # Setup NetQASM arrays and SDK handles.

        # Entanglement results array.
        # This will be filled in by the quantum node controller.
        ent_results_array = self._alloc_ent_results_array(
            number=params.number, tp=EPRType.M
        )

        wait_all = params.post_routine is None

        # Construct and add the NetQASM instructions
        if role == EPRRole.CREATE:
            # NetQASM array for entanglement request parameters.
            create_args_array: Array = self._alloc_epr_create_args(EPRType.M, params)

            self._build_cmds_epr_create_measure(
                create_args_array, ent_results_array, wait_all, params
            )
        else:
            self._build_cmds_epr_recv_measure(ent_results_array, wait_all, params)

        results = deserialize_epr_measure_results(params, ent_results_array, role)
        return results

    def sdk_epr_rsp_create(
        self,
        params: EntRequestParams,
    ) -> List[EprMeasureResult]:
        """Build commands for a 'create remote state preparation' EPR operation
        and return the result futures."""
        self._check_epr_args(tp=EPRType.R, params=params)

        # Setup NetQASM arrays and SDK handles.

        # Entanglement results array.
        # This will be filled in by the quantum node controller.
        ent_results_array = self._alloc_ent_results_array(
            number=params.number, tp=EPRType.R
        )

        # NetQASM array for entanglement request parameters.
        create_args_array: Array = self._alloc_epr_create_args(EPRType.R, params)

        wait_all = params.post_routine is None

        # Construct and add the NetQASM instructions
        self._build_cmds_epr_create_rsp(
            create_args_array, ent_results_array, wait_all, params
        )

        return deserialize_epr_measure_results(
            params, ent_results_array, role=EPRRole.CREATE
        )

    def sdk_epr_rsp_recv(
        self,
        params: EntRequestParams,
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Build commands for a 'receive remote state preparation' EPR operation
        and return the created qubits and result futures."""
        self._check_epr_args(tp=EPRType.R, params=params)

        # Setup NetQASM arrays and SDK handles.

        # Entanglement results array.
        # This will be filled in by the quantum node controller.
        ent_results_array = self._alloc_ent_results_array(
            number=params.number, tp=EPRType.K  # Keep since we are receiving RSP
        )

        qubit_ids_array: Optional[Array] = None

        # SDK handles to result values (Qubit objects).
        qubit_futures = self._get_qubit_futures(
            params.number, params.sequential, ent_results_array
        )
        assert all(isinstance(q, Qubit) for q in qubit_futures)

        # Receivers of R-type requests need an array for IDs for the generated qubits.
        virtual_qubit_ids = [q.qubit_id for q in qubit_futures]
        qubit_ids_array = self.alloc_array(init_values=virtual_qubit_ids)  # type: ignore

        wait_all = params.post_routine is None

        # Construct and add the NetQASM instructions
        self._build_cmds_epr_recv_rsp(
            qubit_ids_array, ent_results_array, wait_all, params
        )

        epr_results = deserialize_epr_keep_results(params, ent_results_array)
        return qubit_futures, epr_results

    def sdk_create_epr_keep(
        self, params: EntRequestParams
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Build commands for a 'create and keep' EPR operation and return the
        created qubits and result futures."""
        if params.min_fidelity_all_at_end is not None:
            # If a min-fidelity constraint is specified, wrap the operation in a loop
            assert params.max_tries is not None
            with self.sdk_new_loop_until_context(params.max_tries) as loop:
                qubits, result_array = self.sdk_epr_keep(
                    role=EPRRole.CREATE, params=params, reset_results_array=True
                )

                results = deserialize_epr_keep_results(params, result_array)

                duration = results[-1].generation_duration
                max_time = NVEprCompiler.get_max_time_for_fidelity(
                    params.min_fidelity_all_at_end
                )
                self._connection._logger.info(
                    f"converting min fidelity constraint of "
                    f"{params.min_fidelity_all_at_end} to a max time of {max_time}"
                )

                loop.set_exit_condition(ValueAtMostConstraint(duration, max_time))

                def cleanup(_: BaseNetQASMConnection):
                    result_array.undefine()
                    # If the request was sequential, each pair has already been
                    # measured and does not need to be freed.
                    # Otherwise: free the qubits.
                    if not params.sequential:
                        for q in qubits:
                            q.free()

                loop.set_cleanup_code(cleanup)

            return qubits, results
        else:
            # otherwise, just do the operation once
            qubits, result_array = self.sdk_epr_keep(role=EPRRole.CREATE, params=params)
            results = deserialize_epr_keep_results(params, result_array)
            return qubits, results

    def sdk_recv_epr_keep(
        self, params: EntRequestParams
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Build commands for a 'receive and keep' EPR operation and return the
        created qubits and result futures."""

        if params.min_fidelity_all_at_end is not None:
            # If a min-fidelity constraint is specified, wrap the operation in a loop
            assert params.max_tries is not None
            with self.sdk_new_loop_until_context(params.max_tries) as loop:
                qubits, result_array = self.sdk_epr_keep(
                    role=EPRRole.RECV, params=params, reset_results_array=True
                )

                results = deserialize_epr_keep_results(params, result_array)

                duration = results[-1].generation_duration
                max_time = NVEprCompiler.get_max_time_for_fidelity(
                    params.min_fidelity_all_at_end
                )
                loop.set_exit_condition(ValueAtMostConstraint(duration, max_time))

                def cleanup(_: BaseNetQASMConnection):
                    result_array.undefine()
                    # If the request was sequential, each pair has already been
                    # measured and does not need to be freed.
                    # Otherwise: free the qubits.
                    if not params.sequential:
                        for q in qubits:
                            q.free()

                loop.set_cleanup_code(cleanup)

            return qubits, results
        else:
            # otherwise, just do the operation once
            qubits, result_array = self.sdk_epr_keep(role=EPRRole.RECV, params=params)
            results = deserialize_epr_keep_results(params, result_array)
            return qubits, results

    def sdk_create_epr_measure(
        self, params: EntRequestParams
    ) -> List[EprMeasureResult]:
        """Build commands for a 'create and measure' EPR operation and return the
        result futures."""
        return self.sdk_epr_measure(role=EPRRole.CREATE, params=params)

    def sdk_recv_epr_measure(self, params: EntRequestParams) -> List[EprMeasureResult]:
        """Build commands for a 'receive and measure' EPR operation and return the
        result futures."""
        return self.sdk_epr_measure(role=EPRRole.RECV, params=params)

    def sdk_create_epr_rsp(self, params: EntRequestParams) -> List[EprMeasureResult]:
        """Build commands for a 'create remote state preperation' EPR operation
        and return the result futures."""
        if params.min_fidelity_all_at_end is not None:
            # If a min-fidelity constraint is specified, wrap the operation in a loop
            assert params.max_tries is not None
            with self.sdk_new_loop_until_context(params.max_tries) as loop:
                results = self.sdk_epr_rsp_create(params=params)
                duration = results[-1].generation_duration
                max_time = NVEprCompiler.get_max_time_for_fidelity(
                    params.min_fidelity_all_at_end
                )
                loop.set_exit_condition(ValueAtMostConstraint(duration, max_time))

            return results
        else:
            # otherwise, just do the operation once
            return self.sdk_epr_rsp_create(params=params)

    def sdk_recv_epr_rsp(
        self, params: EntRequestParams
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Build commands for a 'receive remote state preparation' EPR operation
        and return the created qubits and result futures."""

        if params.min_fidelity_all_at_end is not None:
            # If a min-fidelity constraint is specified, wrap the operation in a loop
            assert params.max_tries is not None
            with self.sdk_new_loop_until_context(params.max_tries) as loop:
                qubits, results = self.sdk_epr_rsp_recv(params=params)
                duration = results[-1].generation_duration
                max_time = NVEprCompiler.get_max_time_for_fidelity(
                    params.min_fidelity_all_at_end
                )
                loop.set_exit_condition(ValueAtMostConstraint(duration, max_time))
            return qubits, results
        else:
            # otherwise, just do the operation once
            return self.sdk_epr_rsp_recv(params=params)

    @contextmanager
    def sdk_loop_context(
        self,
        stop: int,
        start: int = 0,
        step: int = 1,
        loop_register: Optional[Union[operand.Register, str]] = None,
    ) -> Iterator[operand.Register]:
        """Build commands for a 'loop' context and return the context object."""
        try:
            pre_commands = self.subrt_pop_all_pending_commands()
            loop_register_result = self._loop_get_register(loop_register, activate=True)
            yield loop_register_result
        finally:
            body_commands = self.subrt_pop_all_pending_commands()
            self._build_cmds_loop(
                pre_commands=pre_commands,
                body_commands=body_commands,
                stop=stop,
                start=start,
                step=step,
                loop_register=loop_register_result,
            )
            self._mem_mgr.remove_active_register(loop_register_result)

    def sdk_loop_body(
        self,
        body: T_LoopRoutine,
        stop: int,
        start: int = 0,
        step: int = 1,
        loop_register: Optional[Union[operand.Register, str]] = None,
    ) -> None:
        """Build commands for looping the code in the specified body."""
        self._build_cmds_loop_body(body, stop, start, step, loop_register)

    def sdk_if_eq(self, op0: T_CValue, op1: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a == b"""
        self._build_cmds_if_stmt(GenericInstr.BEQ, op0, op1, body)

    def sdk_if_ne(self, op0: T_CValue, op1: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a != b"""
        self._build_cmds_if_stmt(GenericInstr.BNE, op0, op1, body)

    def sdk_if_lt(self, op0: T_CValue, op1: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a < b"""
        self._build_cmds_if_stmt(GenericInstr.BLT, op0, op1, body)

    def sdk_if_ge(self, op0: T_CValue, op1: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a >= b"""
        self._build_cmds_if_stmt(GenericInstr.BGE, op0, op1, body)

    def sdk_if_ez(self, op0: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a == 0"""
        self._build_cmds_if_stmt(GenericInstr.BEZ, op0, op1=None, body=body)

    def sdk_if_nz(self, op0: T_CValue, body: T_BranchRoutine) -> None:
        """An effective if-statement where body is a function executing the clause for a != 0"""
        self._build_cmds_if_stmt(GenericInstr.BNZ, op0, op1=None, body=body)

    def sdk_new_if_context(
        self, condition: GenericInstr, op0: T_CValue, op1: Optional[T_CValue]
    ) -> SdkIfContext:
        """Build commands for an 'if' context and return the context object."""
        id = self._next_context_id
        context = SdkIfContext(
            id=id, builder=self, condition=condition, op0=op0, op1=op1
        )
        self._next_context_id += 1
        return context

    def sdk_new_foreach_context(
        self, array: Array, return_index: bool
    ) -> SdkForEachContext:
        """Build commands for an 'foreach' context and return the context object."""
        id = self._next_context_id
        context = SdkForEachContext(
            id=id, builder=self, array=array, return_index=return_index
        )
        self._next_context_id += 1
        return context

    @contextmanager
    def sdk_new_loop_until_context(
        self, max_iterations: int
    ) -> Iterator[SdkLoopUntilContext]:
        """Build commands for a 'loop_until' context and return the context object."""
        try:
            id = self._next_context_id
            context = SdkLoopUntilContext(
                id=id, builder=self, max_iterations=max_iterations
            )
            self._next_context_id += 1
            loop_register = self._loop_until_context_enter(id)
            reg_future = RegFuture(self._connection, loop_register)
            context.set_loop_register(reg_future)
            yield context
        finally:
            assert context.exit_condition is not None
            self._loop_until_context_exit(
                context_id=id,
                context=context,
            )

    @contextmanager
    def sdk_try_context(
        self,
        max_tries: int = 1,
    ) -> Iterator[None]:
        """Build commands for a 'try' context."""
        try:
            pre_commands = self.subrt_pop_all_pending_commands()
            yield
        finally:
            body_commands = self.subrt_pop_all_pending_commands()
            commands = pre_commands + body_commands
            self.subrt_add_pending_commands(commands)

    @contextmanager
    def sdk_create_epr_context(
        self, params: EntRequestParams
    ) -> Iterator[Tuple[FutureQubit, RegFuture]]:
        """Build commands for an EPR context and return an iterator over
        the EPR qubits and indices created in this context."""
        try:
            (
                pre_commands,
                loop_register,
                ent_results_array,
                output,
                pair,
            ) = self._pre_epr_context(role=EPRRole.CREATE, params=params)
            pair_future = RegFuture(self._connection, pair)
            yield output, pair_future
        finally:
            self._post_epr_context(
                pre_commands=pre_commands,
                number=params.number,
                loop_register=loop_register,
                ent_results_array=ent_results_array,
                pair=pair,
            )
